# Copyright (C) Prizmi, LLC - All Rights Reserved
# Unauthorized copying or use of this file is strictly prohibited and subject to prosecution under applicable laws
# Proprietary and confidential


DEFINITIONS = {
        #############################################################################################################
        #                                                LAYER SECTION                                              #
        #############################################################################################################
        "Linear": {
            "type": "object",
            "additionalProperties": False,
            "required": ["type", "out_features"],
            "properties": {
                "type": {
                    "type": "string",
                    "enum": ["Linear"]
                },
                "out_features": {
                    "type": "integer"
                },
                "bias": {
                    "type": "boolean",
                    "default": True
                }
            }
        },
        "ReLU": {
            "type": "object",
            "additionalProperties": False,
            "required": ["type"],
            "properties": {
                "type": {
                    "type": "string",
                    "enum": ["ReLU"]
                }
            }
        },
        "Dropout": {
            "type": "object",
            "additionalProperties": False,
            "required": ["type"],
            "properties": {
                "type": {
                    "type": "string",
                    "enum": ["Dropout"]
                },
                "p": {
                    "type": "number",
                    "default": 0.5
                }
            }
        },
        "LogSoftmax": {
            "type": "object",
            "additionalProperties": False,
            "required": ["type"],
            "properties": {
                "type": {
                    "type": "string",
                    "enum": ["LogSoftmax"]
                },
                "dim": {
                    "type": ["integer", "null"],
                    "default": 1
                }
            }
        },
        "Softmax": {
            "type": "object",
            "additionalProperties": False,
            "required": ["type"],
            "properties": {
                "type": {
                    "type": "string",
                    "enum": ["Softmax"]
                },
                "dim": {
                    "type": ["integer", "null"],
                    "default": 1
                }
            }
        },
        "Conv2d": {
            "type": "object",
            "additionalProperties": False,
            "required": ["type", "out_channels", "kernel_size"],
            "properties": {
                "type": {
                    "type": "string",
                    "enum": ["Conv2d"]
                },
                "out_channels": {
                    "type": "integer"
                },
                "kernel_size": {
                    "type": "integer"
                },
                "stride": {
                    "type": "integer",
                    "default": 1
                },
                "padding": {
                    "type": "integer",
                    "default": 0
                },
                "dilation": {
                    "type": "integer",
                    "default": 1
                },
                "groups": {
                    "type": "integer",
                    "default": 1
                },
                "bias": {
                    "type": "boolean",
                    "default": True
                },
                "padding_mode": {
                    "type": "string",
                    "default": "zeros"
                },
            }
        },
        "MaxPool2d": {
            "type": "object",
            "additionalProperties": False,
            "required": ["type", "kernel_size"],
            "properties": {
                "type": {
                    "type": "string",
                    "enum": ["MaxPool2d"]
                },
                "kernel_size": {
                    "type": "integer"
                },
                "stride": {
                    "type": ["integer", "null"],
                    "default": None
                },
                "padding": {
                    "type": "integer",
                    "default": 0
                },
                "dilation": {
                    "type": "integer",
                    "default": 0
                },
                "ceil_mode": {
                    "type": "boolean",
                    "default": False
                }
            }
        },
        "BatchNorm1d": {
            "type": "object",
            "additionalProperties": False,
            "required": ["type"],
            "properties": {
                "type": {
                    "type": "string",
                    "enum": ["BatchNorm1d"]
                },
                "eps": {
                    "type": ["number", "null"],
                    "default": 1e-5
                },
                "momentum": {
                    "type": "number",
                    "default": 0.1
                },
                "affine": {
                    "type": "boolean",
                    "default": True
                },
                "track_running_stats": {
                    "type": ["boolean", "null"],
                    "default": True
                }
            }
        },
        #############################################################################################################
        #                                            OPTIMIZER SECTION                                              #
        #############################################################################################################
        "Adam": {
            "type": "object",
            "additionalProperties": False,
            "required": ["type"],
            "properties": {
                "type": {
                    "type": "string",
                    "enum": ["Adam"]
                },
                "lr": {
                    "type": "number",
                    "default": 0.001
                },
                "betas": {
                    "type": "array",
                    "default": [0.9, 0.999]
                },
                "eps": {
                    "type": "number",
                    "default": 1e-8
                },
                "weight_decay": {
                    "type": "number",
                    "default": 0
                },
                "amsgrad": {
                    "type": "boolean",
                    "default": False
                }
            }
        },
        "AdamW": {
            "type": "object",
            "additionalProperties": False,
            "required": ["type"],
            "properties": {
                "type": {
                    "type": "string",
                    "enum": ["AdamW"]
                },
                "lr": {
                    "type": "number",
                    "default": 0.001
                },
                "betas": {
                    "type": "array",
                    "default": [0.9, 0.999]
                },
                "eps": {
                    "type": "number",
                    "default": 1e-8
                },
                "weight_decay": {
                    "type": "number",
                    "default": 0.01
                },
                "amsgrad": {
                    "type": "boolean",
                    "default": False
                }
            }
        },
        "SparseAdam": {
            "type": "object",
            "additionalProperties": False,
            "required": ["type"],
            "properties": {
                "type": {
                    "type": "string",
                    "enum": ["SparseAdam"]
                },
                "lr": {
                    "type": "number",
                    "default": 0.001
                },
                "betas": {
                    "type": "array",
                    "default": [0.9, 0.999]
                },
                "eps": {
                    "type": "number",
                    "default": 1e-8
                }
            }
        },
        "Adamax": {
            "type": "object",
            "additionalProperties": False,
            "required": ["type"],
            "properties": {
                "type": {
                    "type": "string",
                    "enum": ["Adamax"]
                },
                "lr": {
                    "type": "number",
                    "default": 0.002
                },
                "betas": {
                    "type": "array",
                    "default": [0.9, 0.999]
                },
                "eps": {
                    "type": "number",
                    "default": 1e-8
                },
                "weight_decay": {
                    "type": "number",
                    "default": 0
                }
            }
        },
        "ASGD": {
            "type": "object",
            "additionalProperties": False,
            "required": ["type"],
            "properties": {
                "type": {
                    "type": "string",
                    "enum": ["ASGD"]
                },
                "lr": {
                    "type": "number",
                    "default": 0.01
                },
                "lambd": {
                    "type": "number",
                    "default": 0.0001
                },
                "alpha": {
                    "type": "number",
                    "default": 0.75
                },
                "t0": {
                    "type": "number",
                    "default": 1000000.0
                },
                "weight_decay": {
                    "type": "number",
                    "default": 0
                }
            }
        },
        "LBFGS": {
            "type": "object",
            "additionalProperties": False,
            "required": ["type"],
            "properties": {
                "type": {
                    "type": "string",
                    "enum": ["LBFGS"]
                },
                "lr": {
                    "type": "number",
                    "default": 1
                },
                "max_iter": {
                    "type": "integer",
                    "default": 20
                },
                "max_eval": {
                    "type": "integer",
                    "default": 25
                },
                "tolerance_grad": {
                    "type": "number",
                    "default": 1e-05
                },
                "tolerance_change": {
                    "type": "number",
                    "default": 1e-09
                },
                "history_size": {
                    "type": "integer",
                    "default": 100
                },
                "line_search_fn": {
                    "type": ["string", "null"],
                    "default": None
                },
            }
        },
        "RMSprop": {
            "type": "object",
            "additionalProperties": False,
            "required": ["type"],
            "properties": {
                "type": {
                    "type": "string",
                    "enum": ["RMSprop"]
                },
                "lr": {
                    "type": "number",
                    "default": 0.01
                },
                "alpha": {
                    "type": "number",
                    "default": 0.99
                },
                "eps": {
                    "type": "number",
                    "default": 1e-08
                },
                "weight_decay": {
                    "type": "number",
                    "default": 0
                },
                "momentum": {
                    "type": "number",
                    "default": 0
                },
                "centered": {
                    "type": "boolean",
                    "default": False
                }
            }
        },
        "Rprop": {
            "type": "object",
            "additionalProperties": False,
            "required": ["type"],
            "properties": {
                "type": {
                    "type": "string",
                    "enum": ["Rprop"]
                },
                "lr": {
                    "type": "number",
                    "default": 0.01
                },
                "etas": {
                    "type": "array",
                    "default": [0.5, 1.2]
                },
                "step_sizes": {
                    "type": "array",
                    "default": [1e-06, 50]
                }
            }
        },
        "SGD": {
            "type": "object",
            "additionalProperties": False,
            "required": ["type", "lr"],
            "properties": {
                "type": {
                    "type": "string",
                    "enum": ["SGD"]
                },
                "lr": {
                    "type": "number",
                    "default": 0.01
                },
                "momentum": {
                    "type": "number",
                    "default": 0
                },
                "dampening": {
                    "type": "number",
                    "default": 0
                },
                "weight_decay": {
                    "type": "number",
                    "default": 0
                },
                "nesterov": {
                    "type": "boolean",
                    "default": False
                }
            }
        },
        #############################################################################################################
        #                                                LOSS SECTION                                              #
        #############################################################################################################
        "CrossEntropyLoss": {
            "type": "object",
            "additionalProperties": False,
            "required": ["type"],
            "properties": {
                "type": {
                    "type": "string",
                    "enum": ["CrossEntropyLoss"]
                },
                "weight": {
                    "type": ["array", "null"],
                    "default": None
                },
                "ignore_index": {
                    "type": "integer",
                    "default": -100
                },
                "reduction": {
                    "type": "string",
                    "default": "mean"
                }
            }
        },
        "NLLLoss": {
            "type": "object",
            "additionalProperties": False,
            "required": ["type"],
            "properties": {
                "type": {
                    "type": "string",
                    "enum": ["NLLLoss"]
                },
                "weight": {
                    "type": ["array", "null"],
                    "default": None
                },
                "ignore_index": {
                    "type": "integer",
                    "default": -100
                },
                "reduction": {
                    "type": "string",
                    "default": "mean"
                }
            }
        },
        "L1Loss": {
            "type": "object",
            "additionalProperties": False,
            "required": ["type"],
            "properties": {
                "type": {
                    "type": "string",
                    "enum": ["L1Loss"]
                },
                "reduction": {
                    "type": "string",
                    "default": "mean",
                    "enum": ["none", "mean", "sum"]
                }
            }
        },
        "MSELoss": {
            "type": "object",
            "additionalProperties": False,
            "required": ["type"],
            "properties": {
                "type": {
                    "type": "string",
                    "enum": ["MSELoss"]
                },
                "reduction": {
                    "type": "string",
                    "default": "mean",
                    "enum": ["none", "mean", "sum"]
                }
            }
        },
        #############################################################################################################
        #                                                SCHEDULER SECTION                                          #
        #############################################################################################################
        "ExponentialLR": {
            "type": "object",
            "additionalProperties": False,
            "required": ["type", "gamma"],
            "properties": {
                "type": {
                    "type": "string",
                    "enum": ["ExponentialLR"]
                },
                "gamma": {
                    "type": "number",
                    "default": None
                },
                "verbose": {
                    "type": "boolean",
                    "default": False
                }
            }
        },
        "StepLR": {
            "type": "object",
            "additionalProperties": False,
            "required": ["type", "step_size"],
            "properties": {
                "type": {
                    "type": "string",
                    "enum": ["StepLR"]
                },
                "step_size": {
                    "type": "integer"
                },
                "gamma": {
                    "type": "number",
                    "default": 0.1
                },
                "verbose": {
                    "type": "boolean",
                    "default": False
                }
            }
        },
        "MultiStepLR": {
            "type": "object",
            "additionalProperties": False,
            "required": ["type", "milestones"],
            "properties": {
                "type": {
                    "type": "string",
                    "enum": ["MultiStepLR"]
                },
                "milestones": {
                    "type": "array"
                },
                "gamma": {
                    "type": "number",
                    "default": 0.1
                },
                "verbose": {
                    "type": "boolean",
                    "default": False
                }
            }
        },
        "CosineAnnealingLR": {
            "type": "object",
            "additionalProperties": False,
            "required": ["type", "T_max"],
            "properties": {
                "type": {
                    "type": "string",
                    "enum": ["CosineAnnealingLR"]
                },
                "T_max": {
                    "type": "integer"
                },
                "eta_min": {
                    "type": "number",
                    "default": 0
                },
                "verbose": {
                    "type": "boolean",
                    "default": False
                }
            }
        },
        "ReduceLROnPlateau": {
            "type": "object",
            "additionalProperties": False,
            "required": ["type"],
            "properties": {
                "type": {
                    "type": "string",
                    "enum": ["ReduceLROnPlateau"]
                },
                "mode": {
                    "type": "string",
                    "default": "min"
                },
                "factor": {
                    "type": "number",
                    "default": 0.1
                },
                "patience": {
                    "type": "integer",
                    "default": 10
                },
                "threshold": {
                    "type": "number",
                    "default": 0.0001
                },
                "threshold_mode": {
                    "type": "string",
                    "default": "rel"
                },
                "cooldown": {
                    "type": "integer",
                    "default": 0
                },
                "min_lr": {
                    "type": ["number", "array"],
                    "default": 0
                },
                "eps": {
                    "type": "number",
                    "default": 1e-08
                },
                "verbose": {
                    "type": "boolean",
                    "default": False
                }
            }
        },
        "CyclicLR": {
            "type": "object",
            "additionalProperties": False,
            "required": ["type", "base_lr", "max_lr"],
            "properties": {
                "type": {
                    "type": "string",
                    "enum": ["CyclicLR"]
                },
                "base_lr": {
                    "type": ["number", "array"]
                },
                "max_lr": {
                    "type": ["number", "array"]
                },
                "step_size_up": {
                    "type": "integer",
                    "default": 2000
                },
                "step_size_down": {
                    "type": ["null", "integer"],
                    "default": None
                },
                "mode": {
                    "type": "string",
                    "default": "triangular"
                },
                "gamma": {
                    "type": "number",
                    "default": 1.0
                },
                "scale_mode": {
                    "type": "string",
                    "default": "cycle"
                },
                "cycle_momentum": {
                    "type": "boolean",
                    "default": True
                },
                "base_momentum": {
                    "type": ["number", "array"],
                    "default": 0.8
                },
                "max_momentum": {
                    "type": ["number", "array"],
                    "default": 0.9
                },
                "verbose": {
                    "type": "boolean",
                    "default": False
                }
            }
        },
        "OneCycleLR": {
            "type": "object",
            "additionalProperties": False,
            "required": ["type", "max_lr"],
            "properties": {
                "type": {
                    "type": "string",
                    "enum": ["OneCycleLR"]
                },
                "max_lr": {
                    "type": ["number", "array"]
                },
                "total_steps": {
                    "type": ["null", "integer"],
                    "default": None
                },
                "epochs": {
                    "type": ["null", "integer"],
                    "default": None
                },
                "steps_per_epoch": {
                    "type": ["null", "integer"],
                    "default": None
                },
                "pct_start": {
                    "type": "number",
                    "default": 0.3
                },
                "anneal_strategy": {
                    "type": "string",
                    "default": "cos"
                },
                "cycle_momentum": {
                    "type": "boolean",
                    "default": True
                },
                "base_momentum": {
                    "type": ["number", "array"],
                    "default": 0.85
                },
                "max_momentum": {
                    "type": ["number", "array"],
                    "default": 0.95
                },
                "div_factor": {
                    "type": "number",
                    "default": 25.0
                },
                "final_div_factor": {
                    "type": "number",
                    "default": 10000.0
                },
                "three_phase": {
                    "type": "boolean",
                    "default": False
                },
                "verbose": {
                    "type": "boolean",
                    "default": False
                }
            }
        },
        "CosineAnnealingWarmRestarts": {
            "type": "object",
            "additionalProperties": False,
            "required": ["type", "T_0"],
            "properties": {
                "type": {
                    "type": "string",
                    "enum": ["CosineAnnealingWarmRestarts"]
                },
                "T_0": {
                    "type": "integer"
                },
                "T_mult": {
                    "type": "integer",
                    "default": 1
                },
                "eta_min": {
                    "type": "number",
                    "default": 0
                },
                "verbose": {
                    "type": "boolean",
                    "default": False
                }
            }
        },
}


DEEP_CLASSIFIER_PARAMS = {
    "type": "object",
    "default": {},
    "additionalProperties": False,
    "required": ["fully_connected_layers"],
    "properties": {
        "fully_connected_layers": {
            "type": "array",
            "additionalItems": False,
            "minItems": 1,
            "items": {
                "oneOf": [
                    {"$ref": "#/definitions/Linear"},
                    {"$ref": "#/definitions/ReLU"},
                    {"$ref": "#/definitions/Dropout"},
                    {"$ref": "#/definitions/LogSoftmax"},
                    {"$ref": "#/definitions/Softmax"},
                    {"$ref": "#/definitions/BatchNorm1d"},
                ]
            }
        },
        "embedding_dropout": {
            "type": "number",
            "default": 0.5,
            "minimum": 0,
            "maximum": 1
        },
        "epochs": {
            "type": "integer",
            "default": 1,
            "minimum": 1,
        },
        "batch_size": {
            "type": "integer",
            "additionalItems": False,
            "default": 64,
            "minimum": 1,
        },
        "save_interval": {
            "type": "integer",
            "additionalItems": False,
            "default": 2,
            "minimum": 1,
        },
        "embedding_sizes": {
            "type": ["object", "null"],
            "default": None,
        },
        "y_range": {
            "type": ["array", "null"],
            "additionalItems": False,
            "default": None,
            "minItems": 2,
            "maxItems": 2,
            "items": {
                "type": "number"
            }
        },
        "loss": {
            "type": "object",
            "additionalItems": False,
            "default": {"type": "NLLLoss", "weight": None, "ignore_index": -100, "reduction": 'mean'},
            "oneOf": [
                {"$ref": "#/definitions/CrossEntropyLoss"},
                {"$ref": "#/definitions/NLLLoss"}
            ]
        },
        "scheduler": {
            "type": "object",
            "additionalItems": False,
            "oneOf": [
                {"$ref": "#/definitions/StepLR"},
                {"$ref": "#/definitions/MultiStepLR"},
                {"$ref": "#/definitions/CosineAnnealingLR"},
                {"$ref": "#/definitions/ExponentialLR"},
                {"$ref": "#/definitions/ReduceLROnPlateau"},
                {"$ref": "#/definitions/CyclicLR"},
                {"$ref": "#/definitions/CosineAnnealingWarmRestarts"},
            ]
        },
        "optimizer": {
            "type": "object",
            "additionalItems": False,
            "default": {"type": "Adam", "lr": 0.001, 'betas': [0.9, 0.999], 'eps': 1e-08, 'weight_decay': 0,
                        'amsgrad': False},
            "oneOf": [
                {"$ref": "#/definitions/Adam"},
                {"$ref": "#/definitions/AdamW"},
                {"$ref": "#/definitions/SparseAdam"},
                {"$ref": "#/definitions/Adamax"},
                {"$ref": "#/definitions/ASGD"},
                {"$ref": "#/definitions/LBFGS"},
                {"$ref": "#/definitions/RMSprop"},
                {"$ref": "#/definitions/Rprop"},
                {"$ref": "#/definitions/SGD"},
            ]
        },
    },
    "definitions": DEFINITIONS
}


DEEP_REGRESSOR_PARAMS = {
    "type": "object",
    "default": {},
    "additionalProperties": False,
    "required": ["fully_connected_layers"],
    "properties": {
        "fully_connected_layers": {
            "type": "array",
            "additionalItems": False,
            "minItems": 1,
            "items": {
                "oneOf": [
                    {"$ref": "#/definitions/Linear"},
                    {"$ref": "#/definitions/ReLU"},
                    {"$ref": "#/definitions/Dropout"},
                    {"$ref": "#/definitions/BatchNorm1d"},
                ]
            }
        },
        "embedding_dropout": {
            "type": "number",
            "default": 0.5,
            "minimum": 0,
            "maximum": 1
        },
        "epochs": {
            "type": "integer",
            "default": 1,
            "minimum": 1,
        },
        "batch_size": {
            "type": "integer",
            "additionalItems": False,
            "default": 64,
            "minimum": 1,
        },
        "save_interval": {
            "type": "integer",
            "additionalItems": False,
            "default": 2,
            "minimum": 1,
        },
        "embedding_sizes": {
            "type": ["object", "null"],
            "default": None,
        },
        "y_range": {
            "type": ["array", "null"],
            "additionalItems": False,
            "default": None,
            "minItems": 2,
            "maxItems": 2,
            "items": {
                "type": "number"
            }
        },
        "loss": {
            "type": "object",
            "additionalItems": False,
            "default": {"type": "MSELoss", "reduction": "mean"},
            "oneOf": [
                {"$ref": "#/definitions/L1Loss"},
                {"$ref": "#/definitions/MSELoss"},
            ]
        },
        "scheduler": {
            "type": "object",
            "additionalItems": False,
            "oneOf": [
                {"$ref": "#/definitions/StepLR"},
                {"$ref": "#/definitions/MultiStepLR"},
                {"$ref": "#/definitions/CosineAnnealingLR"},
                {"$ref": "#/definitions/ExponentialLR"},
                {"$ref": "#/definitions/ReduceLROnPlateau"},
                {"$ref": "#/definitions/CyclicLR"},
                {"$ref": "#/definitions/CosineAnnealingWarmRestarts"},
            ]
        },
        "optimizer": {
            "type": "object",
            "additionalItems": False,
            "default": {"type": "Adam", "lr": 0.001, 'betas': [0.9, 0.999], 'eps': 1e-08, 'weight_decay': 0,
                        'amsgrad': False},
            "oneOf": [
                {"$ref": "#/definitions/Adam"},
                {"$ref": "#/definitions/AdamW"},
                {"$ref": "#/definitions/SparseAdam"},
                {"$ref": "#/definitions/Adamax"},
                {"$ref": "#/definitions/ASGD"},
                {"$ref": "#/definitions/LBFGS"},
                {"$ref": "#/definitions/RMSprop"},
                {"$ref": "#/definitions/Rprop"},
                {"$ref": "#/definitions/SGD"},
            ]
        },
    },
    "definitions": DEFINITIONS
}


IMAGE_CLASSIFIER_PARAMS = {
    "type": "object",
    "default": {},
    "additionalProperties": False,
    "required": ["classifier"],
    "properties": {
        "classifier": {
            "type": "array",
            "additionalItems": False,
            "minItems": 1,
            "items": {
                "oneOf": [
                    {"$ref": "#/definitions/Linear"},
                    {"$ref": "#/definitions/ReLU"},
                    {"$ref": "#/definitions/Dropout"},
                    {"$ref": "#/definitions/LogSoftmax"},
                    {"$ref": "#/definitions/Softmax"},
                ]
            }
        },
        "features": {
            "type": "array",
            "additionalItems": False,
            "minItems": 1,
            "items": {
                "oneOf": [
                    {"$ref": "#/definitions/Conv2d"},
                    {"$ref": "#/definitions/ReLU"},
                    {"$ref": "#/definitions/MaxPool2d"}
                ]
            }
        },
        "freeze_features": {
            "type": "boolean",
            "default": True,
        },
        "loss": {
            "type": "object",
            "additionalItems": False,
            "default": {"type": "NLLLoss", "weight": None, "ignore_index": -100, "reduction": 'mean'},
            "oneOf": [
                {"$ref": "#/definitions/CrossEntropyLoss"},
                {"$ref": "#/definitions/NLLLoss"},
            ]
        },
        "scheduler": {
            "type": "object",
            "additionalItems": False,
            "oneOf": [
                {"$ref": "#/definitions/StepLR"},
                {"$ref": "#/definitions/MultiStepLR"},
                {"$ref": "#/definitions/CosineAnnealingLR"},
                {"$ref": "#/definitions/ExponentialLR"},
                {"$ref": "#/definitions/ReduceLROnPlateau"},
                {"$ref": "#/definitions/CyclicLR"},
                {"$ref": "#/definitions/CosineAnnealingWarmRestarts"},
            ]
        },
        "optimizer": {
            "type": "object",
            "additionalItems": False,
            "default": {"type": "Adam", "lr": 0.001, 'betas': [0.9, 0.999], 'eps': 1e-08, 'weight_decay': 0, 'amsgrad': False},
            "oneOf": [
                {"$ref": "#/definitions/Adam"},
                {"$ref": "#/definitions/AdamW"},
                {"$ref": "#/definitions/SparseAdam"},
                {"$ref": "#/definitions/Adamax"},
                {"$ref": "#/definitions/ASGD"},
                {"$ref": "#/definitions/LBFGS"},
                {"$ref": "#/definitions/RMSprop"},
                {"$ref": "#/definitions/Rprop"},
                {"$ref": "#/definitions/SGD"},
            ]
        },
        "epochs": {
            "type": "integer",
            "additionalItems": False,
            "default": 1,
            "minimum": 1,
        },
        "save_interval": {
            "type": "integer",
            "additionalItems": False,
            "default": 2,
            "minimum": 1,
        },
        "batch_size": {
            "type": "integer",
            "additionalItems": False,
            "default": 64,
            "minimum": 1,
        }
    },
    "definitions": DEFINITIONS
}


IMAGE_TRANSFORM_PARAMS = {
    "type": "array",
    "additionalItems": False,
    "minItems": 1,
    "items": {
        "oneOf": [
            {
                "type": "object",
                "additionalProperties": False,
                "required": ["type", "degrees"],
                "properties": {
                     "type": {
                         "type": "string",
                         "enum": ["RandomRotation"]
                     },
                     "activate": {
                         "type": "boolean",
                         "default": True
                     },
                     "degrees": {
                         "type": ["number", "array"]
                     },
                     "expand": {
                        "type": "boolean",
                        "default": False
                     },
                     "center": {
                        "type": ["null", "array"],
                        "default": None
                     },
                     "fill": {
                        "type": ["number", "array"],
                        "default": 0
                     },
                }
            },
            {
                "type": "object",
                "additionalProperties": False,
                "required": ["type"],
                "properties": {
                    "type": {
                        "type": "string",
                        "enum": ["RandomHorizontalFlip"]
                    },
                    "activate": {
                        "type": "boolean",
                        "default": True
                    },
                    "p": {
                        "type": "number",
                        "default": 0.5
                    },
                }
            },
            {
                "type": "object",
                "additionalProperties": False,
                "required": ["type", "size"],
                "properties": {
                    "type": {
                        "type": "string",
                        "enum": ["Resize"]
                    },
                    "activate": {
                        "type": "boolean",
                        "default": True
                    },
                    "size": {
                        "type": ["integer", "array"]
                    },
                }
            },
            {
                "type": "object",
                "additionalProperties": False,
                "required": ["type", "size"],
                "properties": {
                    "type": {
                        "type": "string",
                        "enum": ["CenterCrop"]
                    },
                    "activate": {
                        "type": "boolean",
                        "default": True
                    },
                    "size": {
                        "type": ["integer", "array"]
                    },
                }
            },
            {
                "type": "object",
                "additionalProperties": False,
                "required": ["type"],
                "properties": {
                    "type": {
                        "type": "string",
                        "enum": ["ToTensor"]
                    },
                    "activate": {
                        "type": "boolean",
                        "default": True
                    }
                }
            },
            {
                "type": "object",
                "additionalProperties": False,
                "required": ["type", "mean", "std"],
                "properties": {
                    "type": {
                        "type": "string",
                        "enum": ["Normalize"]
                    },
                    "activate": {
                        "type": "boolean",
                        "default": True
                    },
                    "mean": {
                        "type": "array"
                    },
                    "std": {
                        "type": "array"
                    },
                }
            },
        ]
    }
}




# Copyright (C) Prizmi, LLC - All Rights Reserved
# Unauthorized copying or use of this file is strictly prohibited and subject to prosecution under applicable laws
# Proprietary and confidential

"""The main module for parsing the instructions from the configuration file."""

import sys
import logging
from pprint import pformat

from learner.setup.setup import Setup
from learner.configuration.supported_items import *
from learner.configuration.data import DataConfiguration
from learner.configuration.column import ColumnConfiguration
from learner.configuration.process import ProcessConfiguration
from learner.configuration.validation import ValidationConfiguration
from learner.configuration.workspace import WorkspaceConfiguration
from learner.configuration.model import ModelsConfiguration
from learner.configuration.segmenter import SegmenterConfiguration
from learner.configuration.recommender import RecommenderConfiguration
from learner.configuration.similarities import SimilaritiesConfiguration
from learner.configuration.outlier import OutlierConfiguration
from learner.configuration.feature_engineering import FeatureEngineeringConfiguration
from learner.configuration.sample import SampleConfiguration
from learner.configuration.analysis import AnalysisConfiguration
from learner.configuration.connection import ConnectionConfiguration
from learner.configuration.communication import CommunicationConfiguration
from learner.configuration.combine import CombineConfiguration
from learner.validator.column_validator import ColumnValidator
from learner.validator.model_validator import ModelValidator


class Configuration(Setup):
    """Parse the json configuration and make assignments for all input variables that Learner supports. Some variables
    may not be defined in the configuration file. In that case, Learner uses default values and throws an error if that
    variable is not optional. Each section of the configuration fils is handled by a nested class."""

    def __init__(self, json_config_file=None):
        super().__init__(json_config_file)
        self.log_configuration_file()
        self.engine = self.get_engine()
        self.workspace = WorkspaceConfiguration(self._json_config, self.output_path, self.verbose_level, self.output_name)
        self.sample = SampleConfiguration(self._json_config, self.engine, self.workspace)
        self.data = DataConfiguration(self._json_config, self.workspace, self.sample, self.engine, self.sep_timetag)
        self.column = ColumnConfiguration(self._json_config, self.engine, self.data, self.sample)
        self.process = ProcessConfiguration(self._json_config, self.data, self.column, self.sample, self.engine)
        self.validation = ValidationConfiguration(self._json_config)
        self.segmenter = SegmenterConfiguration(self._json_config, self.process, self.column)
        self.model = ModelsConfiguration(self._json_config, self.sep_timetag, self.workspace, self.data, self.engine, self.segmenter)
        self.recommender = RecommenderConfiguration(self._json_config)
        self.similarities = SimilaritiesConfiguration(self._json_config)
        self.outlier = OutlierConfiguration(self._json_config, self.column, self.data, self.process)
        self.feature_engineering = FeatureEngineeringConfiguration(self._json_config, self.column, self.process)
        self.analysis = AnalysisConfiguration(self._json_config, self.column, self.workspace, self.data, self.sep_timetag, self.engine)
        self.connection = ConnectionConfiguration(self._json_config, self.data)
        self.communication = CommunicationConfiguration(self._json_config, self.connection.credentials)
        self.combine = CombineConfiguration(self._json_config, self.model, self.sep_timetag, self.workspace, self.engine, self.data)
        column_validator = ColumnValidator(self)
        column_validator.validate_columns()
        model_validator = ModelValidator(self)
        model_validator.validate_models()
        logging.info("Successfully prepared the configuration...")

    def get_engine(self):
        # remove the underscore
        engine = "".join(self._json_config["engine"].split("_"))
        # make sure the engine name is case-insensitive
        for eng in SUPPORTED_ENGINES:
            if engine.lower() == eng.lower():
                engine = eng
        if engine not in SUPPORTED_ENGINES:
            logging.error("Invalid engine, supported engines are %s. Exiting...", SUPPORTED_ENGINES)
            sys.exit(1)
        return engine

    def log_configuration_file(self):
        logging.info(f"---Printing the content of the configuration file: {self.config_filename} ---")
        # we do not want to log the license key for security reasons
        self._json_config.pop("license_key", self._json_config)
        logging.info(pformat(self._json_config))
        logging.info("---Finished printing the content of the configuration file---")

    def __getstate__(self):
        # Copy the object's state from self.__dict__ which contains
        # all our instance attributes. Always use the dict.copy()
        # method to avoid modifying the original state.
        state = self.__dict__.copy()
        # Remove the unpicklable entries.
        del state['connection']
        return state



# Copyright (C) Prizmi, LLC - All Rights Reserved
# Unauthorized copying or use of this file is strictly prohibited and subject to prosecution under applicable laws
# Proprietary and confidential

"""The main module for combining the predictions of multiple models and creating an ensemble model."""
import os
import logging

from scipy import stats
import pandas as pd
import numpy as np
import pickle

from learner.model_manager.scoring_manager import BaseScorer
from learner.data_worker.data_loader import get_data
from learner.configuration.configuration import Configuration
from learner.model_manager.prediction_manager import BasePredictor
from learner.model_manager.scoring_manager import RegressorScorer, ClassifierScorer
from learner.utilities.templates import PRED_PATH, SAVE_MODEL_PATH


class MeanRegressorCombiner:
    """This class handles combining the regression models. In regression models, only one column is needed for
    predictions but in classification model, multiple columns may be needed for predictions (multiclass classification).
    As such, two different classes handle these models.
    """
    def __init__(self, conf: Configuration, models_dict, sep_timetag, data_type):
        """Initialize an instance of RegressorCombiner using a conf object, models_dict, and sep_timetag.

        :param conf: an instance of the Configuration class
        :param models_dict: learner's models dictionary. This dictionary contains the information related to all models.
        :param sep_timetag: separator plus timetag or nothing (if we shouldn't use the timetag in the output files)
        :param data_type: the data type, it can be "validation" or "test"
        """
        self._conf = conf
        self._models_dict = models_dict
        self._sep_timetag = sep_timetag
        # this will hold the id_cols df if defined
        self._id_cols = None
        # this holds all the prediction data
        self._pred = pd.DataFrame()
        self.data_type = data_type
        self.dtype_sep = f"_{self.data_type}_"
        self.column_name = getattr(self._conf.data, f"{self.data_type}_column_name")
        self.add_timetag = getattr(self.conf.data, f"{self.data_type}_add_timetag")

    @property
    def conf(self):
        return self._conf

    @property
    def models_dict(self):
        return self._models_dict

    @property
    def sep_timetag(self):
        return self._sep_timetag

    @property
    def pred(self):
        return self._pred

    def combine(self):
        """Loop through each model in models_dict, load the prediction (and id) columns, and then combine the
        predictions. After than, write the predictions to disk and score them if requested.

        :return: None
        """
        for tag, mdls in self._models_dict.items():
            mdl = mdls.get(0, mdls)
            logging.info("Loading the predictions for model %s", tag)
            filename = PRED_PATH.format(
                path=mdl["path"],
                output_name=self._conf.workspace.name,
                dtype_sep=self.dtype_sep,
                tag=str(tag),
                sep_timetag=str(self._conf.sep_timetag)
            )
            if self._id_cols is None:
                self.load_id_cols(filename)

            pred = self.load_pred(tag, filename)
            self._pred = pd.concat([self._pred, pred], axis=1)

        # compute the average of all predictions
        if self.conf.combine.mean_type == "arithmetic":
            self._pred[self.column_name] = self._pred.mean(axis=1)
        # if the mean type is not arithmetic, it's geometric
        else:
            self._pred[self.column_name] = stats.gmean(self.pred, axis=1)

        combine_dict = self.get_mean_combine_dict()
        predictor = BasePredictor(None, self._conf)
        predictor.write_predictions(next(iter(combine_dict)),
                                    self._pred[self.column_name],
                                    data_type=self.data_type,
                                    data=self._id_cols,
                                    models_dict=combine_dict,
                                    add_timetag=self.add_timetag)

        # we can score the combined model if requested
        if self.data_type == "validation" and self._conf.data.validation_score_activate:
            scorer = RegressorScorer(self._conf, combine_dict)
            scorer.score()

    def load_id_cols(self, filename):
        """Load the id columns from one of the prediction files. This method should be called once because the id
        columns should be identical for all models.

        :param filename: the path to the output (prediction file)
        :return: None
        """
        if self._conf.column.id_cols:
            self._id_cols = get_data(filename,
                                     manifest_file=None,
                                     usecols=self._conf.column.id_cols)

    def load_pred(self, tag, filename):
        """Load the predictions for a single model. After the data is loaded, rename the prediction column to be the
        model tag.

        :param tag: the model tag (this is used for renaming the prediction column)
        :param filename: the path to the prediction file
        :return: a pandas dataframe containing the prediction data
        """
        logging.info("Loading prediction column from the output file...")

        pred = get_data(filename,
                        manifest_file=None,
                        usecols=[self.column_name]).rename(columns={self.column_name: tag})
        return pred

    def get_mean_combine_dict(self):
        """Construct a models_dict items for mean_combine model. The models_dict for the combined models is constructed
        here instead of configuration module because this model cannot be loaded, etc.

        :return: A models_dict with the tag "mean_combine".
        """
        combine_dict = {}
        tag = f"mean_combine"
        combine_dict[tag] = {}

        directory_path = self._conf.workspace.path + self._conf.workspace.name + "_" + tag + str(self._sep_timetag) + "/"

        if not os.path.exists(directory_path):  # pragma: no cover
            os.makedirs(directory_path)
        combine_dict[tag]["path"] = directory_path

        return combine_dict


class MeanClassifierCombiner:
    """This class handles combining the classification models. In multiclass classification models where the
    probabilities are predicted, the prediction file will contain multiple prediction columns unlike the regression
    models where only one column is needed for predictions. As such, two different classes handle the classification
    and regression models.
    """
    def __init__(self, conf: Configuration, models_dict, sep_timetag, data_type):
        """Initialize an instance of ClassifierCombiner using a conf object, models_dict, and sep_timetag.

        :param conf: an instance of the Configuration class
        :param models_dict: learner's models dictionary. This dictionary contains the information related to all models.
        :param sep_timetag: separator plus timetag or nothing (if we shouldn't use the timetag in the output files)
        :param data_type: the data type, it can be "validation" or "test"
        """
        self._conf = conf
        self._models_dict = models_dict
        self._sep_timetag = sep_timetag
        # this will hold the id_cols df if defined
        self._id_cols = None
        # pandas dataframe to hold the prediction for class and probabilities
        self._pred_class = pd.DataFrame()
        self._pred_proba = pd.DataFrame()
        # these will hold the column names for class and proba
        self._proba_col_names = None
        self._class_col_names = None
        self.validation_pred_cols = None
        self.test_pred_cols = None
        self.data_type = data_type
        self.dtype_sep = f"_{self.data_type}_"
        self.prediction_type = getattr(self._conf.data, f"{self.data_type}_prediction_type")
        self.add_timetag = getattr(self.conf.data, f"{self.data_type}_add_timetag")

    @property
    def conf(self):
        return self._conf

    @property
    def models_dict(self):
        return self._models_dict

    @property
    def sep_timetag(self):
        return self._sep_timetag

    @property
    def pred_class(self):
        return self._pred_class

    @property
    def pred_proba(self):
        return self._pred_proba

    @property
    def proba_col_names(self):
        return self._proba_col_names

    @property
    def class_col_names(self):
        return self._class_col_names

    def combine(self):
        """Depending on the prediction type (proba, class, or all), call the corresponding methods to combine the
        prediction of the models. After that, write the combined predictions to disk and score them if needed.

        :return: None
        """
        pred = pd.DataFrame()
        if self.prediction_type in ["all", "proba"]:
            self.combine_proba()
            pred = pd.concat([pred, self._pred_proba[self._proba_col_names]], axis=1)
        if self.prediction_type in ["all", "class"]:
            self.combine_class()
            pred = pd.concat([pred, self._pred_class[self._class_col_names]], axis=1)

        combine_dict = self.get_mean_combine_dict()
        predictor = BasePredictor(None, self._conf)
        predictor.write_predictions(next(iter(combine_dict)),
                                    pred,
                                    data_type=self.data_type,
                                    data=self._id_cols,
                                    models_dict=combine_dict,
                                    add_timetag=self.add_timetag)

        # we can score the combined model if requested
        if self.conf.engine in ["Classifier", "Regressor"] and self._conf.data.validation_score_activate:
            scorer = ClassifierScorer(self._conf, combine_dict)
            scorer.score()

    def combine_proba(self):
        """Loop through each model in models_dict, load the prediction (and id) columns, and then combine the
        predictions.

        :return: None
        """
        for tag, mdls in self._models_dict.items():
            mdl = mdls.get(0, mdls)
            # we need this to form the models_dict
            if getattr(self, f"{self.data_type}_pred_cols") is None:
                setattr(self, f"{self.data_type}_pred_cols", mdl[f"{self.data_type}_pred_cols"])
            filename = PRED_PATH.format(
                path=mdl["path"],
                output_name=self._conf.workspace.name,
                dtype_sep=self.dtype_sep,
                tag=str(tag),
                sep_timetag=str(self._conf.sep_timetag)
            )

            # if we haven't loaded the id columns, do it here
            if self._id_cols is None:
                self.load_id_cols(filename)

            logging.info("Loading the predictions for model %s", tag)
            pred = self.load_pred_proba(tag, mdl, filename)
            self._pred_proba = pd.concat([self._pred_proba, pred], axis=1)

        # this dictionary stores the classes (keys) and the corresponding columns for those classes. This is then used
        # to compute the averages for each class
        class_cols = {}
        for col in self._proba_col_names:
            c = col.rsplit("_")[-1]
            class_cols[c] = [col for col in self._pred_proba if col.endswith(c)]
            if self.conf.combine.mean_type == "arithmetic":
                self._pred_proba[col] = self._pred_proba[class_cols[c]].mean(axis=1)
            else:
                self._pred_proba[col] = stats.gmean(self._pred_proba[class_cols[c]], axis=1)

    def combine_class(self):
        """Loop through each model in models_dict, load the prediction (and id) columns, and then combine the
        predictions.

        :return: None
        """
        for tag, mdls in self._models_dict.items():
            mdl = mdls.get(0, mdls)
            # we need this to form the models_dict
            if getattr(self, f"{self.data_type}_pred_cols") is None:
                setattr(self, f"{self.data_type}_pred_cols", mdl[f"{self.data_type}_pred_cols"])
            logging.info("Loading the predictions for model %s", tag)
            filename = PRED_PATH.format(
                path=mdl["path"],
                output_name=self._conf.workspace.name,
                dtype_sep=self.dtype_sep,
                tag=str(tag),
                sep_timetag=str(self._conf.sep_timetag)
            )

            # if we haven't loaded the id columns, do it here
            if self._id_cols is None:
                self.load_id_cols(filename)

            pred = self.load_pred_class(tag, mdl, filename)
            self._pred_class = pd.concat([self._pred_class, pred], axis=1)

        # to determine the predicted class, we do majority vote here
        self._pred_class[self._class_col_names] = self._pred_class.mode(axis=1).iloc[:, 0].astype(int)

    def load_id_cols(self, filename):
        """Load the id columns from one of the prediction files. This method should be called once because the id
        columns should be identical for all models.

        :param filename: the path to the output (prediction file)
        :return: None
        """
        if self._conf.column.id_cols:
            self._id_cols = get_data(filename,
                                     manifest_file=None,
                                     usecols=self._conf.column.id_cols)

    def load_pred_proba(self, tag, mdl, filename):
        """Load the predictions for a single model. After the data is loaded, prepend the models tag to the name of the
        columns in the prediction file.

        :param tag: the model tag (this is used for renaming the prediction column)
        :param filename: the path to the prediction file
        :return: a pandas dataframe containing the prediction data
        """
        logging.info("Loading prediction probabilities from the output file...")

        # get the prediction column names
        if self.prediction_type == "all":
            self._proba_col_names = list(filter(lambda x: x.startswith('_proba_'), mdl[f"{self.data_type}_pred_cols"]))
        else:
            self._proba_col_names = list(mdl[f"{self.data_type}_pred_cols"])

        # load the prediction columns
        pred_proba = get_data(filename,
                              manifest_file=None,
                              usecols=self._proba_col_names)
        # rename the column name to start with the model tag
        col_rename_dict = {col_name: f"{tag}_{col_name}" for col_name in pred_proba.columns}
        pred_proba = pred_proba.rename(columns=col_rename_dict)
        return pred_proba

    def load_pred_class(self, tag, mdl, filename):
        """Load the predictions for a single model. After the data is loaded, rename the prediction column to be the
        model tag.

        :param tag: the model tag (this is used for renaming the prediction column)
        :param filename: the path to the prediction file
        :return: a pandas dataframe containing the prediction data
        """
        logging.info("Loading prediction column from the output file...")

        if self.prediction_type == "all":
            self._class_col_names = list(filter(lambda x: x.startswith('_class_'), mdl[f"{self.data_type}_pred_cols"]))
        else:
            self._class_col_names = list(mdl[f"{self.data_type}_pred_cols"])

        pred_class = get_data(filename,
                              manifest_file=None,
                              usecols=self._class_col_names)

        # rename the column name to start with the model tag
        col_rename_dict = {col_name: f"{tag}" for col_name in pred_class.columns}
        # class column name should be only one item
        self._class_col_names = self._class_col_names[0]
        pred_class = pred_class.rename(columns=col_rename_dict)
        return pred_class

    def get_mean_combine_dict(self):
        """Construct a models_dict items for mean_combine model. The models_dict for the combined models is constructed
        here instead of configuration module because this model cannot be loaded, etc.

        :return: A models_dict with the tag "mean_combine".
        """
        combine_dict = {}
        tag = "mean_combine"
        combine_dict[tag] = {}

        directory_path = self._conf.workspace.path + self._conf.workspace.name + "_" + tag + str(self._sep_timetag) + "/"

        if not os.path.exists(directory_path):  # pragma: no cover
            os.makedirs(directory_path)
        combine_dict[tag]["path"] = directory_path
        combine_dict[tag][f"{self.data_type}_pred_cols"] = getattr(self, f"{self.data_type}_pred_cols")

        return combine_dict


class TriadCombiner(BaseScorer):
    """This class implements the TriadCombiner algorithm to combine the predictions of multiple models. For more details
    about the algorithm see <link to be added after once the blog post is uploaded>. TriadCombiner is only supported for
    regression problems (both standard and deep regression). This class follows the fit-transform pattern. Either fit or
    transform is called. The "fit" method fits the weights and the "transform" method uses the weights to get the final
    predictions. Please note that the "fit" method also writes the final predictions for validation data.
    """
    def __init__(self, conf: Configuration, models_dict, sep_timetag):
        """Initialize an instance of TriadCombiner using a conf object, models_dict, and sep_timetag.

        :param conf: an instance of the Configuration class
        :param models_dict: learner's models dictionary. This dictionary contains the information related to all models.
        :param sep_timetag: separator plus timetag or nothing (if we shouldn't use the timetag in the output files)
        """
        super().__init__(conf, models_dict)
        self._sep_timetag = sep_timetag
        # these are the initial weights. We currently use 0.5 but this can potentially come from a distribution
        self.w = np.zeros(2 * self._conf.model.num_models - 2)
        self.w.fill(0.5)
        self.n_triads = self._conf.model.num_models - 1
        # this is list of initial scores for each model
        self.initial_scores = []
        self.best_initial_score = None
        self.best_final_score = None
        # the following lists would hold models and triads weights
        self.models_weights = []
        self.triad_weights = []
        # this will hold the id_cols df if defined
        self._id_cols = None
        self.data_type = None
        self.dtype_sep = None
        self.column_name = None
        self.add_timetag = None

    @property
    def conf(self):
        return self._conf

    @conf.setter
    def conf(self, value):
        self._conf = value

    @property
    def models_dict(self):
        return self._models_dict

    @property
    def sep_timetag(self):
        return self._sep_timetag

    def fit(self):
        """The "fit" method for TriadCombiner. This method first update the instance attributes to correspond to
        "validation" data_type. It then loads the predictions (on validation data) to fit the parameters. Two sets of
        parameters need to be determined: i) model weights, which multiplies each models predictions by a constant,
        ii) the triad weights. Finally, it writes the combined predictions and saves the combiner.

        :return: None
        """
        self.update_params_for_data_type(data_type="validation")
        pred, id_cols = self.load_pred_id_cols()
        # this will save all of the nodes data
        nodes = np.zeros((pred.shape[0], 2*self.conf.model.num_models - 1))
        nodes = self.get_model_weights(pred, nodes)
        self.get_triad_weights(nodes)
        self.calculate_total_improvement()
        combine_dict = self.get_triad_combine_dict()
        self.write_predictions(id_cols, nodes, combine_dict)
        self.save_combiner(combine_dict)

    def transform(self, data_type):
        """The "transform" method for TriadCombiner. The assumption here is that the model_weights and triad_weights
        are already populated. This method accepts a data_type argument. This method first update the instance
        attributes to correspond to the data_type argument. It then loads the predictions (validation or test data
        depending on the value of data_type argument. Finally, it writes the combined predictions.

        :param data_type: the data_type, it can be "validation" or "test"
        :return: None
        """
        self.update_params_for_data_type(data_type=data_type)
        pred, id_cols = self.load_pred_id_cols()
        nodes = np.zeros((pred.shape[0], 2*self.conf.model.num_models - 1))
        for i in range(self._conf.model.num_models):
            nodes[:, i] = pred[:, i] * self.models_weights[i]
        for triad in range(self.n_triads):
            nodes[:, self._conf.model.num_models + triad] = self.triad_weights[2*triad] * nodes[:, 2*triad] + self.triad_weights[2*triad+1] * nodes[:, 2*triad+1]
        combine_dict = self.get_triad_combine_dict()
        self.write_predictions(id_cols, nodes, combine_dict)

    def update_params_for_data_type(self, data_type):
        """Update the data_type related instance attributes. The reason for updating the instance attributes here rather
        than the init method is that the combiner can be used to transform both validation and test data without
        re-instantiating the combiner object.

        :param data_type: the data_type, it can be "validation" or "test"
        :return: None
        """
        self.data_type = data_type
        self.dtype_sep = f"_{data_type}_"
        self.column_name = getattr(self._conf.data, f"{data_type}_column_name")
        self.add_timetag = getattr(self.conf.data, f"{data_type}_add_timetag")

    def load_pred_id_cols(self):
        """Load the predictions and id data. This method loops through all models defined in models_dict and loads the
        predictions one by one and appends them to the pred dataframe. It also loads the id columns once because id
        columns are identical for all models. As the predictions of different models are loaded, it also scores the
        predictions to get the best initial score. This best initial score is later used to get the total improvement
        from combining the predictions of multiple models.

        :return: pred 2d numpy array and id_cols dataframe
        """
        pred = pd.DataFrame()
        self.initial_scores = []
        id_cols = None
        for tag, mdls in self._models_dict.items():
            mdl = mdls.get(0, mdls)
            logging.info("Loading the predictions for model %s", tag)
            filename = PRED_PATH.format(
                path=mdl["path"],
                output_name=self._conf.workspace.name,
                dtype_sep=self.dtype_sep,
                tag=str(tag),
                sep_timetag=str(self._conf.sep_timetag)
            )
            if id_cols is None:
                id_cols = self.load_id_cols(filename)

            single_pred = self.load_pred(tag, filename)
            self.initial_scores.append(self.scoring_handler(self.conf.combine.triad_score_type, self.y_true, single_pred, params={}))
            pred = pd.concat([pred, single_pred], axis=1)

        pred = pred.values
        self.best_initial_score = min(self.initial_scores)
        return pred, id_cols

    def load_id_cols(self, filename):
        """Load the id columns from one of the prediction files. This method should be called once because the id
        columns should be identical for all models.

        :param filename: the path to the output (prediction file)
        :return: None
        """
        if self._conf.column.id_cols:
            id_cols = get_data(filename,
                               manifest_file=None,
                               usecols=self._conf.column.id_cols)
            return id_cols

    def load_pred(self, tag, filename):
        """Load the predictions for a single model. After the data is loaded, rename the prediction column to be the
        model tag.

        :param tag: the model tag (this is used for renaming the prediction column)
        :param filename: the path to the prediction file
        :return: a pandas dataframe containing the prediction data
        """
        logging.info("Loading prediction column from the output file...")

        pred = get_data(filename,
                        manifest_file=None,
                        usecols=[self.column_name]).rename(columns={self.column_name: tag})
        return pred

    def get_model_weights(self, pred, nodes):
        """In the triad algorithms, we multiply the predictions of each model by a constant number. In this method,
        those weights are determined and the nodes array is updated with the weighted predictions.

        :param pred: a 2-d numpy array containing the predictions
        :param nodes: a 2-d numpy array to hold all the nodes data.
        :return: updated nodes data where the first layer contains the weighted predictions.
        """
        for triad_index in range(self.conf.model.num_models):
            nodes[:, triad_index], weight, s = self.single_weight_gradient_descent(0.5,
                                                                                   pred[:, triad_index],
                                                                                   self.y_true,
                                                                                   dw=self.conf.combine.triad_models_step_size,
                                                                                   max_iter=self.conf.combine.triad_models_max_iter)
            self.models_weights.append(weight)
        return nodes

    def single_weight_gradient_descent(self, w, pred, y_true, dw, max_iter):
        """Find the optimum weight to multiply the predictions of a model given the predictions and true values using
        gradient descent algorithms.

        :param w: the initial weight, typically 0.5.
        :param pred: a prediction column
        :param y_true: the true values
        :param dw: the step size for gradient descent
        :param max_iter: the maximum number of iterations before we sto[
        :return: the weithed optimum predictions, the optimum weight, and the score
        """
        counter = 1
        s_old, old_prediction = self.get_weighted_score(w, pred, y_true)
        s_new, new_prediction = self.get_weighted_score(w + dw, pred, y_true)

        if s_new < s_old:
            sign = +1
        elif s_new == s_old:
            return new_prediction, w, s_new
        else:
            sign = -1

        delta_s = dw  # dw and delta_s must positive
        while counter < max_iter:

            w += sign * delta_s
            s_new, new_prediction = self.get_weighted_score(w, pred, y_true)

            if s_new > s_old:
                w -= sign * delta_s
                break
            else:
                old_prediction = new_prediction

            s_old = s_new
            counter += 1
            if counter % 10000 == 0:
                logging.info(f"haven't converged after {counter} counts, new score is {s_new} and weight is {w}")

        return old_prediction, w, s_old

    def get_weighted_score(self, w, pred, y_true):
        """Multiply the predictions of a model by a wight and compute the score.

        :param w: the weight
        :param pred: an array of predictions
        :param y_true: an array of true values
        :return: the score the weighted predictions
        """
        pred = w * pred
        score = self.scoring_handler(self.conf.combine.triad_score_type, y_true, pred, params={})

        return score, pred

    def get_triad_weights(self, nodes):
        """This is one of the main methods for the triad algorithms. Here we run the gradient descent algorithms to
        find the optimum weight for each triad.

        :param nodes: a 2-d numpy array to hold all the nodes data
        :return: None
        """
        initial_scores = self.get_model_scores(self.conf.model.num_models, self.y_true, nodes)
        logging.info(f"Initial scores for individual models are {initial_scores}")

        for triad_index in range(self.n_triads):
            w1, w2, s = self.double_weight_gradient_descent(self.w[2 * triad_index],
                                                            self.w[2 * triad_index + 1],
                                                            nodes,
                                                            triad_index,
                                                            self.y_true,
                                                            dw=self.conf.combine.triad_triads_step_size,
                                                            max_iter=self.conf.combine.triad_triads_max_iter,
                                                            n_models=self.conf.model.num_models)
            logging.info("w1={w1:.4f} \t w2={w2:.4f} \t s={s:.6f} \t triad index = {triad_index}".
                         format(w1=w1, w2=w2, s=s, triad_index=triad_index))
            self.triad_weights.append(w1)
            self.triad_weights.append(w2)
            self.best_final_score = s

    def get_model_scores(self, num_models, y_true, pred):
        """Compute the scores of all models. This will help us get the best initial score. Please note that the score
        type is defined for the triad algorithm. The pred is a 2-d numpy array that contains all predictions.

        :param num_models: the number of models, it's used to loop over the columns of pred array
        :param y_true: the true values to use for scoring
        :param pred: an 2-d numpy array that contains all the predictions
        :return: a 1-d numpy array containing the scores for all the models
        """
        s0 = np.zeros(num_models)
        for i in range(num_models):
            s0[i] = self.scoring_handler(self.conf.combine.triad_score_type, y_true, pred[:, i], params={})
        return s0

    def double_weight_gradient_descent(self, w1, w2, pred, triad_index, y_true, dw, max_iter, n_models):
        """This method runs the gradient descent algorithms to optimize the weights for combining the predictions of
        two models in a triad geometry.

        :param w1: the initial first weight
        :param w2: the initial second weight
        :param pred: a 2-d numpy array containing all the predictions
        :param triad_index: the index of the triad, 0, 1, 2, ...
        :param y_true: the true values
        :param dw: the step size for the gradient descent algorithm
        :param max_iter: the maximum number of iterations before we stop
        :param n_models: the total number of models
        :return: the optimum weights the the corresponding score
        """
        counter = 1
        s_old, old_pred = self.get_combined_score(w1, w2, y_true, pred[:, 2 * triad_index], pred[:, 2 * triad_index + 1])
        s_new, new_pred = self.get_combined_score(w1 + dw, w2 - dw, y_true, pred[:, 2 * triad_index], pred[:, 2 * triad_index + 1])

        if s_new < s_old:
            sign = +1
        elif s_new == s_old:
            return w1, w2, s_new
        else:
            sign = -1

        delta_s = dw  # dw and delta_s must positive
        while counter < max_iter:

            w1 += sign * delta_s
            w2 -= sign * delta_s
            s_new, new_pred = self.get_combined_score(w1, w2, y_true, pred[:, 2 * triad_index], pred[:, 2 * triad_index + 1])

            if s_new > s_old:
                w1 -= sign * delta_s
                w2 += sign * delta_s
                break
            else:
                old_pred = new_pred

            counter += 1
            s_old = s_new

            if counter % 10000 == 0:
                logging.info(f"haven't converged after {counter} counts, new score is {s_new} and wights are {w1}, {w2}")

        if triad_index + n_models < pred.shape[1]:
            pred[:, triad_index + n_models] = old_pred
        return w1, w2, s_old

    def get_combined_score(self, w1, w2, y_true, pred1, pred2):
        """Use two weights to combine the predictions of two models and then compute the score of the combined models.

        :param w1: the first weight to multiply the predictions of the first model
        :param w2: the second weight to multiply the predictions of the second model
        :param y_true: the true values
        :param pred1: the predictions of the first model
        :param pred2: the predictions of the second model
        :return: the calculated score and the combined predictions
        """
        pred = w1 * pred1 + w2 * pred2
        s = self.scoring_handler(self.conf.combine.triad_score_type, y_true, pred, params={})
        return s, pred

    def calculate_total_improvement(self):
        """Use the best initial scire and the final score to compute the percentage of the improvement.

        :return: None
        """
        percentage_improvement = ((self.best_initial_score - self.best_final_score) / self.best_initial_score) * 100
        logging.info("Total improvement={:.4f}%".format(percentage_improvement))

    def write_predictions(self, id_cols, nodes, combine_dict):
        """Write the final predictions after combining all the models into disk similar to all other models.

        :param id_cols: a pandas dataframe containing the id columns
        :param nodes: a 2-d numpy array containing all the nodes data. The last column has the final predictions.
        :param combine_dict: a models dict for triad combine algorithm
        :return: None
        """
        predictor = BasePredictor(None, self._conf)
        predictor.write_predictions(next(iter(combine_dict)),
                                    pd.Series(nodes[:, -1], name=self.column_name),
                                    data_type=self.data_type,
                                    data=id_cols,
                                    models_dict=combine_dict,
                                    add_timetag=self.add_timetag)

    def save_combiner(self, combine_dict):
        """Save/pickle the combiner object so it can be loaded if needed. Please note that we only save the combiner
        object if we are training and we want to save the models.

        :param combine_dict: a models dict for triad combine algorithm
        :return: None
        """
        tag = next(iter(combine_dict))
        # save the combiner if train_models save_models are set to true
        if self.conf.model.train_models and self._conf.model.save_models:
            filename = SAVE_MODEL_PATH.format(
                path=combine_dict[tag]["path"],
                output_name=self._conf.workspace.name,
                tag=tag,
                sep_timetag=str(self.sep_timetag),
                ext=".triad")

            # we need to do this because we cannot directly pickle the deep learning models
            for tag, mdl in self.models_dict.items():
                mdl["model"] = None
            with open(filename, "wb") as pickle_file:
                pickle.dump(self, pickle_file)

    def get_triad_combine_dict(self):
        """Construct a models_dict items for tria_combine model. The models_dict for the combined models is constructed
        here instead of configuration module because this model cannot be loaded like other models.

        :return: A models_dict with the tag "triad_combine".
        """
        combine_dict = {}
        tag = "triad_combine"
        combine_dict[tag] = {}

        directory_path = self._conf.workspace.path + self._conf.workspace.name + "_" + tag + str(self._sep_timetag) + "/"

        if not os.path.exists(directory_path):  # pragma: no cover
            os.makedirs(directory_path)
        combine_dict[tag]["path"] = directory_path

        return combine_dict



# Copyright (C) Prizmi, LLC - All Rights Reserved
# Unauthorized copying or use of this file is strictly prohibited and subject to prosecution under applicable laws
# Proprietary and confidential

import sys
import logging

from learner.data_worker.data_loader import get_value
from learner.validator.input_validator import validate_subset_list
from learner.configuration.supported_items import SUPPORTED_ENGINES_FOR_TRIAD_COMBINE, SUPPORTED_SCORE_TYPES, \
    SUPPORTED_MEAN_COMBINE_TYPES, SUPPORTED_ENGINES_FOR_MEAN_COMBINE


class CombineConfiguration:
    """Parse all input variables related to combining the predictions of multiple models."""

    def __init__(self, json_config, model, sep_timetag, output, engine, data):
        self._json_config = json_config
        self._model = model
        self.sep_timetag = sep_timetag
        self._output = output
        self._engine = engine
        self._data = data
        self.mean_activate = self.get_mean_activate()
        self.mean_type = self.get_mean_type()

        self.triad_activate = self.get_triad_activate()
        self.triad_score_type = self.get_triad_score_type()
        self.triad_models_step_size = get_value(self._json_config, 0.01, "combine", "triad_params", "models_step_size")
        self.triad_models_max_iter = get_value(self._json_config, 1000, "combine", "triad_params", "models_max_iter")
        self.triad_triads_step_size = get_value(self._json_config, 0.001, "combine", "triad_params", "triads_step_size")
        self.triad_triads_max_iter = get_value(self._json_config, 10000, "combine", "triad_params", "triads_max_iter")

    def get_mean_activate(self):
        try:
            activate = self._json_config["combine"]["mean_params"]["activate"]
            # if we only have one model, activating combine won't make sense.
            if activate:
                validate_subset_list(parent_list=SUPPORTED_ENGINES_FOR_MEAN_COMBINE,
                                     parent_name="acceptable engines for mean combining method",
                                     subset_list=[self._engine],
                                     subset_name="the defined engine")

                if self._model.train_models and self._model.num_models == 1:
                    logging.critical("Combine functionality was activated but only one model is defined. Please "
                                     "update your configuration file and try again. Exiting...")
                    sys.exit(1)
            return activate
        except KeyError:
            return False

    def get_mean_type(self):
        try:
            mean_type = self._json_config["combine"]["mean_params"]["type"]
            validate_subset_list(parent_list=SUPPORTED_MEAN_COMBINE_TYPES,
                                 parent_name="supported mean types for combining the models",
                                 subset_list=[mean_type],
                                 subset_name="the defined mean type")
            return mean_type
        except KeyError:
            return "arithmetic"

    def get_triad_activate(self):
        try:
            activate = self._json_config["combine"]["triad_params"]["activate"]
            if activate:
                validate_subset_list(parent_list=SUPPORTED_ENGINES_FOR_TRIAD_COMBINE,
                                     parent_name="acceptable engines for triad combining method",
                                     subset_list=[self._engine],
                                     subset_name="the defined engine")

                # we need to activate scoring_params is we are training the models and doing triad combine
                if self._model.train_models and not self._data.validation_score_activate:
                    self._data.validation_score_activate = True

                # if we only have one model, activating combine won't make sense.
                # (n & (n - 1)) == 0 is bit manipulation to check if n is a power of 2
                if self._model.num_models == 1 or not (self._model.num_models & (self._model.num_models - 1)) == 0:
                    logging.critical(f"The number of models should be a multiple of 2 when activating triad combine "
                                     f"functionality but you defined {self._model.num_models} models. Please "
                                     f"update your configuration file and try again. Exiting...")
                    sys.exit(1)
            return activate
        except KeyError:
            return False

    def get_triad_score_type(self):
        try:
            score_type = self._json_config["combine"]["triad_params"]["score_type"]
            if self.triad_activate:
                if score_type not in SUPPORTED_SCORE_TYPES[self._engine]:
                    logging.error("The score_type defined in triad_params is invalid, supported score types for %s engine "
                                  "are %s. Exiting...", self._engine, sorted(SUPPORTED_SCORE_TYPES[self._engine]))
                    sys.exit(1)
            return score_type
        except KeyError:
            return "root_mean_squared_error"



# Copyright (C) Prizmi, LLC - All Rights Reserved
# Unauthorized copying or use of this file is strictly prohibited and subject to prosecution under applicable laws
# Proprietary and confidential

"""This module implements the ModelValidator class to validate the model parameters provided by users.
"""
import sys
import logging

from learner.configuration.defaults import DEEP_CLASSIFIER_PARAMS, DEEP_REGRESSOR_PARAMS, IMAGE_CLASSIFIER_PARAMS
from jsonschema import Draft7Validator, validators


def extend_with_default(validator_class):
    validate_properties = validator_class.VALIDATORS["properties"]

    def set_defaults(validator, properties, instance, schema):
        for property, subschema in properties.items():
            if "type" in instance and "type" in properties and \
               instance["type"] == properties["type"]["enum"][0] and \
               "default" in subschema:
                instance.setdefault(property, subschema["default"])
            if "type" not in instance and "default" in subschema:
                instance.setdefault(property, subschema["default"])
        for error in validate_properties(validator, properties, instance, schema,):
            yield error

    return validators.extend(validator_class, {"properties": set_defaults},)


DefaultValidatingDraft7Validator = extend_with_default(Draft7Validator)


class ModelValidator:
    """Validate the model parameters defined in the configuration file. This class only validates the custom models
    such as DeepClassifier and DeepRegressor. We currently do not validate standard models such as RandomForest, etc.
    To understand what models are validates, look at DEFAULT_MODEL_PARAMS
    """
    def __init__(self, conf):
        self.conf = conf

    def validate_models(self):
        """The main method that communicates with other methods to validate the model parameters. The name of each
        validator method ends with the "type" of the model.

        :return: None
        """
        self.validate_deep_classifier()
        self.validate_deep_regressor()
        self.validate_image_classifier()

    def validate_deep_classifier(self):
        for tag, mdl in self.conf.model.models_dict.items():
            # we currently only need to do the validation for certain type of models. Segmentation is not supported
            # for these models. As such, we are limiting validations to only dict types
            if not isinstance(mdl, dict):
                return
            if mdl["type"] == "DeepClassifier":
                DefaultValidatingDraft7Validator(DEEP_CLASSIFIER_PARAMS).validate(mdl["params"])
                # make sure the loss function is consistent with the last layer
                if mdl["params"]["fully_connected_layers"][-1]["type"] == "LogSoftmax" and mdl["params"]["loss"]["type"] == "CrossEntropyLoss":
                    logging.critical("You last fully connected layer is LogSoftmax and you are using CrossEntropyLoss. "
                                     "The CrossEntropyLoss already includes LogSoftmax for you. Please delete that "
                                     "layer or use NLLLoss instead. Exiting...")
                    sys.exit(1)

    def validate_deep_regressor(self):
        for tag, mdl in self.conf.model.models_dict.items():
            # we currently only need to do the validation for certain type of models. Segmentation is not supported
            # for these models. As such, we are limiting validations to only dict types
            if not isinstance(mdl, dict):
                return
            if mdl["type"] == "DeepRegressor":
                DefaultValidatingDraft7Validator(DEEP_REGRESSOR_PARAMS).validate(mdl["params"])
                # we need to make sure the last Linear layer has 1 node
                i = -1
                while abs(i) <= len(mdl["params"]["fully_connected_layers"]):
                    if mdl["params"]["fully_connected_layers"][i]["type"] == "Linear":
                        if mdl["params"]["fully_connected_layers"][i]["out_features"] != 1:
                            logging.critical("In deep_regressor engine, the last Linear layer should have only 1 node, "
                                             "i.e. out_features: 1. Please update the configuration file and try "
                                             "again. Existing...")
                            sys.exit(1)
                        break
                    i -= 1

    def validate_image_classifier(self):
        if self.conf.engine != "ImageClassifier":
            return
        for tag, mdl in self.conf.model.models_dict.items():
            # we currently only need to do the validation for certain type of models. Segmentation is not supported
            # for these models. As such, we are limiting validations to only dict types
            if not isinstance(mdl, dict):
                return
            DefaultValidatingDraft7Validator(IMAGE_CLASSIFIER_PARAMS).validate(mdl["params"])
            # make sure the loss function is consistent with the last layer
            if mdl["params"]["classifier"][-1]["type"] == "LogSoftmax" and mdl["params"]["loss"]["type"] == "CrossEntropyLoss":
                logging.critical("You last classifier layer is LogSoftmax and you are using CrossEntropyLoss. The "
                                 "CrossEntropyLoss already includes LogSoftmax for you. Please delete that layer or "
                                 "use NLLLoss instead. Exiting...")
                sys.exit(1)
            # make sure "stride" is set to "kernel_size" in MaxPool2d if it is undefined, i.e. None
            if "features" in mdl["params"]:
                for layer in mdl["params"]["features"]:
                    if layer["type"] == "MaxPool2d" and layer["stride"] is None:
                        layer["stride"] = layer["kernel_size"]
            # TODO: convert the weight parameter in NLLLoss to a tensor



# Copyright (C) Prizmi, LLC - All Rights Reserved
# Unauthorized copying or use of this file is strictly prohibited and subject to prosecution under applicable laws
# Proprietary and confidential

"""Authenticate to Google API and download data from google drive."""
import os
import sys
import io
import pickle
import logging
from googleapiclient.discovery import build
from google_auth_oauthlib.flow import InstalledAppFlow
from google.auth.transport.requests import Request
from googleapiclient.http import MediaIoBaseDownload, MediaFileUpload
from google.oauth2 import service_account

logging.getLogger("googleapiclient").setLevel(logging.CRITICAL)
# If modifying these scopes, delete the file token.pickle.
SCOPES = ['https://www.googleapis.com/auth/drive', 'https://www.googleapis.com/auth/sqlservice.admin']


class BaseGoogleDrive:  # pragma: no cover
    """The base class for google driver reader (and writer) classes. This class implements the methods that the child
    classes share. These are methods like authentication, etc.
    """
    def __init__(self, write_token=True, keys=None):
        """Initialize the object using the sheet id, range and a flag that defines whether a token should be written
        onto the disk or not.

        :param write_token: a boolean flag that determines if the refresh token should be stored on disk or not. This is
        useful in situation where we don't have permission to write or don't want to write the refresh token for other
        reasons.
        """
        self.write_token = write_token
        self.keys = keys
        self.service = self.authenticate()

    def authenticate(self):
        """Authenticate to Google API.

        :return: an authenticated build object
        """
        if self.keys:
            creds = service_account.Credentials.from_service_account_info(
                self.keys, scopes=SCOPES)
        else:
            creds = None
            # The file token.pickle stores the user's access and refresh tokens, and is
            # created automatically when the authorization flow completes for the first
            # time.
            path_to_token = os.path.join(os.path.dirname(__file__), "..", "credentials", 'gdrive_token.pickle')
            if os.path.exists(path_to_token):
                with open(path_to_token, 'rb') as token:
                    creds = pickle.load(token)
            # If there are no (valid) credentials available, let the user log in.
            if not creds or not creds.valid:
                if creds and creds.expired and creds.refresh_token:
                    creds.refresh(Request())
                else:
                    path_to_credentials = os.path.join(os.path.dirname(__file__), "..", "credentials", '.gdrive_credentials.json')
                    flow = InstalledAppFlow.from_client_secrets_file(path_to_credentials, SCOPES)
                    # the default port may not work here
                    creds = flow.run_local_server(port=8899)
                # Save the credentials for the next run
                if self.write_token:
                    with open(path_to_token, 'wb') as token:
                        pickle.dump(creds, token)

        service = build('drive', 'v3', credentials=creds, cache_discovery=False)
        return service

    def get_file_name(self, id):
        """Given a file id, return the name of that file. This is useful in situations where we know the file id and
        we want to preserve the filename when downloading it.

        :param id: the id of the file on google drive
        :return: the string filenameit
        """
        # populate the filename attribute in case one may need to use it
        file = self.service.files().get(fileId=id).execute()
        return file["name"]

    @staticmethod
    def get_url(id):
        """Construct complete google drive url from id.

        :param id: id for google drive object.
        :type id: string
        :return: full url for google drive object.
        :rtype: string
        """
        return "https://drive.google.com/open?id={}".format(id)

    def get_file_owner(self, id):
        """Accept a file id and return the email address of the file owner

        :param id: id for google drive object.
        :return: the email address of the file owner
        """
        try:
            permissions = self.service.permissions().list(fileId=id).execute()
        except Exception:
            logging.critical("Something went wrong when trying to obtain the owner of the file. Exiting...")
            sys.exit(1)
        permission_items = permissions.get('items', [])
        for item in permission_items:
            if item["role"] == "owner":
                return item["emailAddress"]
        logging.critical("No owner was found for the file. Exiting...")
        sys.exit(1)

    @staticmethod
    def _call_back(request_id, response, exception):
        """Call back function used in sending request to google drive api for logging information.

        :param request_id: placeholder. needed in order to be correctly passed to `new_batch_http_request`
        :param response: http call result from request.
        :param exception: Exception that would cause warning log
        :return: None
        """
        if exception:
            # Handle error
            logging.warning(exception)
        else:
            logging.info("Permission Id: {}".format(response.get('id')))


class GoogleDriveReader(BaseGoogleDrive):  # pragma: no cover
    """Download files and obtain various information about the files stored on google drive. Information such as
    file_id, file_name, etc"""

    def __init__(self, write_token=True, keys=None):
        super().__init__(write_token, keys)
        self.values = None

    def read_file_get_handle(self, id):
        """Read a file (identified by its id) and get a file handler. The file handler can then be used to write the
        data to disk. The file handler has methods such as get_values(), seek(), etc, which enables navigation and other
        operations.

        :param id: the id of the file on google drive
        :return: a file handler (BytesIO object)
        """
        request = self.service.files().export_media(fileId=id,
                                                    mimeType='text/plain')
        fh = io.BytesIO()
        downloader = MediaIoBaseDownload(fh, request)
        done = False
        while done is False:
            status, done = downloader.next_chunk()
            logging.info("Download %d%%." % int(status.progress() * 100))

        return fh

    def get_file_id(self, filename):
        """Given a filename on google drive, try to search and find the file id. If there are multiple file stored with
        the filename, raise an error.

        :param filename: the name of the file to be searched on google drive
        :return: the id of the file on google drive
        """
        results = self.service.files().list(pageSize=10,
                                            spaces='drive',
                                            fields="nextPageToken, files(id, name)",
                                            q="name = '{filename}'".format(filename=filename)).execute()
        item = results.get('files', [])
        if not item:
            logging.error('No files found.')
        elif len(item) > 1:
            logging.error("More than one file found using the name {filename}. Update the name or use a different "
                          "method if this is the expected behavior. Exiting...".format(filename=filename))
        else:
            return item[0]['id']

    def download_to_file(self, id, filename):
        """Read a file (identified by its id) and get a file handler. The file handler can then be used to write the
        data to disk. The file handler has methods such as get_values(), seek(), etc, which enables navigation and other
        operations.

        :param id: the id of the file on google drive
        :params filename: the full path (path + name) to be saved on disk
        :return: a file handler (BytesIO object)
        """
        request = self.service.files().get_media(fileId=id)
        fh = io.FileIO(filename, 'wb')
        downloader = MediaIoBaseDownload(fh, request)
        done = False
        while done is False:
            status, done = downloader.next_chunk()
            logging.info("Download %d%%." % int(status.progress() * 100))


class GoogleDriveWriter(BaseGoogleDrive):  # pragma: no cover
    """Upload files and obtain various information about the files stored on google drive. Information such as
        file_id, file_name, etc
    """

    def __init__(self, write_token=True, keys=None):
        super().__init__(write_token, keys)
        self.values = None

    def write_file(self, filename, name=None, mimetype=None, resumable=True):
        """Given a filename on local machine, try to upload the file into google drive, and return the id of the
        uploaded file

        :param filename: the name of the file to be uploaded to google drive
        :type filename: string
        :param name: title of file to be shown in google drive. If None, will use same filename. Default is None.
        :type name: None or string
        :param mimetype: Mime-type of the file. If None then a mime-type will be guessed from the file extension.
        :type mimetype: string
        :param resumable: whether to use resumable upload or not. For large files, use True. Default is True.
        :type resumable: bool
        :return: the id of the file on google drive
        """
        file_metadata = {'name': name if name is not None else os.path.split(filename)[-1]}
        media = MediaFileUpload(filename,
                                mimetype=mimetype,
                                resumable=resumable)
        file = self.service.files().create(body=file_metadata,
                                           media_body=media,
                                           fields='id').execute()
        return file.get("id")

    def delete_file(self, id):
        """Delete the file with the given id

        :param id: id of the file on google drive
        :type id: string
        :return: return the name of the deleted file
        :rtype: string
        """
        filename = self.get_file_name(id=id)
        self.service.files().delete(fileId=id).execute()
        return filename



# Copyright (C) Prizmi, LLC - All Rights Reserved
# Unauthorized copying or use of this file is strictly prohibited and subject to prosecution under applicable laws
# Proprietary and confidential

"""This module implements classes and functions for building the layers for neural network models. Currently, this
module implements the LayerBuilder class. In the future, we'll most likely implement a parent class and let more
specialized classes inherit from it."""
import sys
from copy import deepcopy
import logging

import torch.nn as nn

from learner.data_worker.data_processor import delete_keys_from_dict


class LayerBuilder:
    """The class to build a list of neural network layers using a "raw" list that contains the necessary information
    and parameters for building those layers. The main method here is build_layers. This method calls other methods to
    build the layers"""
    def __init__(self, raw_list, initial_size, height=100, width=100):
        """Initialize a LayerBuilder object using the input parameters.

        :param raw_list: a list that contains all the necessary parameters to construct the layers
        :param initial_size: the size for building the layers. This is typically the size of input layer
        :param height: the height of the images, if applicable
        :param width: the width of the images, if applicable
        """
        self._initial_size = initial_size
        self.running_size = initial_size
        self.height = height
        self.width = width
        self.running_flatten_size = initial_size * height * width
        self._raw_list = raw_list

    @property
    def initial_size(self):
        return self._initial_size

    @property
    def raw_list(self):
        return self._raw_list

    def build_layers(self):
        """The main method for building the list of the layers. Here, we iterate through the raw list and using
        functional programming, we build one layer at a time.

        :return:
        """
        layer_list = []
        # make a copy because we may alter the list and want to keep the original list intact
        raw_list = deepcopy(self.raw_list)
        try:
            for i, layer in enumerate(raw_list):
                layer_list.append(getattr(self, f"build_{layer['type'].lower()}")(i, layer))
            return layer_list
        except Exception as e:
            logging.critical(f"Unable to build the layers. The error is {e}")
            sys.exit(1)

    def build_linear(self, index, layer):
        """Build a Linear layer. If we are building the first layer, use the initial size to set in_features parameter,
        otherwise use the running_size. Before returning the layer, set the running_size to out_features.

        :param index: the index of the layer, 0, 1, 2, ...
        :param layer: a dictionary that contains all the parameters.
        :return: a Linear object with the proper parameters
        """
        # if we are building the first layer, we need to know the initial size. If not, our input
        # is our running size
        if index == 0:
            in_features = self.initial_size
        else:
            in_features = self.running_size
        self.running_size = layer["out_features"]
        delete_keys_from_dict(layer, ["type"])
        return nn.Linear(in_features=in_features, **layer)

    def build_relu(self, index, layer):
        """Build a ReLu activation layer.

        :param index: the index of the layer, 0, 1, 2, .... This is not used here.
        :param layer: a dictionary that contains all the parameters. This is not used here.
        :return: a ReLu object
        """
        return nn.ReLU()

    def build_softmax(self, index, layer):
        """Build a Softmax activation layer.

        :param index: the index of the layer, 0, 1, 2, .... This is not used here.
        :param layer: a dictionary that contains all the parameters.
        :return: a Softmax object
        """
        delete_keys_from_dict(layer, ["type"])
        return nn.Softmax(**layer)

    def build_logsoftmax(self, index, layer):
        """Build a LogSoftmax activation layer.

        :param index: the index of the layer, 0, 1, 2, .... This is not used here.
        :param layer: a dictionary that contains all the parameters.
        :return: a LogSoftmax object
        """
        delete_keys_from_dict(layer, ["type"])
        return nn.LogSoftmax(**layer)

    @staticmethod
    def build_dropout(index, layer):
        """Build a dropout layer.

        :param index: the index of the layer, 0, 1, 2, .... This is not used here.
        :param layer: a dictionary that contains all the parameters.
        :return: a Dropout object
        """
        delete_keys_from_dict(layer, ["type"])
        return nn.Dropout(**layer)

    def build_conv2d(self, index, layer):
        """Build a Conv2d layer. If we are at the first layer, we use the initial_size to set the in_channel. If not,
        the running_size is our in_channel. Once, the layer is built, out_channel will be our running_size. We also,
        use the running_size, height, and width to get the running_flatten_size. This value can be used when connecting
        the convolutional layers to the fully connected layers. This method also calculates the height and width of the
        output images. This way, the user won't have to figure out those numbers. This commonly is a source of
        frustration for many people.

        :param index: the index of the layer, 0, 1, 2, ...
        :param layer: a dictionary that contains all the parameters.
        :return: a Conv2d object
        """
        # if we are building the first layer, we need to know the initial size. If not, our input
        # is our running size
        if index == 0:
            in_channels = self.initial_size
        else:
            in_channels = self.running_size

        self.running_size = layer["out_channels"]
        self.height = int((self.height + 2 * layer["padding"] - layer["dilation"] * (layer["kernel_size"] - 1) - 1) / layer["stride"] +1)
        self.width = int((self.width + 2 * layer["padding"] - layer["dilation"] * (layer["kernel_size"] - 1) - 1) / layer["stride"] +1)
        self.running_flatten_size = self.running_size * self.height * self.width
        delete_keys_from_dict(layer, ["type"])
        return nn.Conv2d(in_channels=in_channels, **layer)

    def build_maxpool2d(self, index, layer):
        """Build a MaxPool2d layer. We first compute the height and the width of the output images. We then use those
        numbers to compute the running_flatten_size. We then return the layer.

        :param index: the index of the layer, 0, 1, 2, ...
        :param layer: a dictionary that contains all the parameters.
        :return: a MaxPool2d object.
        """
        self.height = int((self.height + 2 * layer["padding"] - layer["dilation"] * (layer["kernel_size"] - 1) - 1) / layer["stride"] +1)
        self.width = int((self.width + 2 * layer["padding"] - layer["dilation"] * (layer["kernel_size"] - 1) - 1) / layer["stride"] +1)
        self.running_flatten_size = int(self.running_size * self.height * self.width)
        delete_keys_from_dict(layer, ["type"])
        return nn.MaxPool2d(**layer)

    def build_batchnorm1d(self, index, layer):
        """Build a BatchNorm1d layer. We set the num_features to the running_size. This layer does not change the
        running_size.

        :param index: the index of the layer, 0, 1, 2, ...
        :param layer: a dictionary that contains all the parameters.
        :return: a BatchNorm1d object with the proper parameters
        """
        delete_keys_from_dict(layer, ["type"])
        return nn.BatchNorm1d(num_features=self.running_size, **layer)



# Copyright (C) Prizmi, LLC - All Rights Reserved
# Unauthorized copying or use of this file is strictly prohibited and subject to prosecution under applicable laws
# Proprietary and confidential

"""The main module to drive the feature engineering operations."""

import logging
import warnings
import numpy as np

from learner.data_worker.data_processor import DataProcessor


class FeatureEngineering:
    """The main feature engineering class that implements the methods for performing feature engineering tasks."""

    def __init__(self, conf):
        self._conf = conf
        self._processor = None
        self._learner_groupby = None

    @property
    def conf(self):
        return self._conf

    @property
    def processor(self):
        return self._processor

    @property
    def learner_groupby(self):
        return self._learner_groupby

    def basic_operations(self, data):
        """Loop through each item in basic_operations and make the call to appropriate methods to perform basic
        operations feature engineering including addition, subtraction, multiplication, and division.

        :param data: the data frame
        :return: the updated data frame
        """
        for item in self._conf.feature_engineering.basic_operations_params:
            data = getattr(self, item["method"])(item, data)
        return data

    def division(self, item, data):
        """Divide a sequence of columns to create a new feature.

        :param item: an item in basic_operations_params list
        :param data: the data frame
        :return: the updated data frame
        """
        data[item["name"]] = data[item["cols"][0]]
        if "value" in item:
            try:
                data[item["name"]] = (data[item["name"]] / item["value"]).replace(np.inf, np.nan)
            except Exception as e:
                warnings.warn("The division operation failed when dividing col {col} by {value}. The error is: {error}".
                              format(col=item["cols"][0], value=item["value"], error=str(e)), Warning)
                self._conf.feature_engineering.drop_cols.append(item["name"])

        for col in item["cols"][1:]:
            try:
                data[item["name"]] = (data[item["name"]] / data[col]).replace(np.inf, np.nan)
            except Exception as e:
                warnings.warn("The division operation failed while processing {0}. The error is: {1}".
                              format(col, str(e)), Warning)
                self._conf.feature_engineering.drop_cols.append(item["name"])
        return data

    def multiplication(self, item, data):
        """Multiply a sequence of columns to create a new feature.

        :param item: an item in basic_operations_params list
        :param data: the data frame
        :return: the updated data frame
        """
        data[item["name"]] = data[item["cols"][0]]
        if "value" in item:
            try:
                data[item["name"]] = data[item["name"]] * item["value"]
            except Exception as e:
                warnings.warn("The multiplication operation failed when multiplying col {col} by {value}. "
                              "The error is: {error}".
                              format(col=item["cols"][0], value=item["value"], error=str(e)), Warning)
                self._conf.feature_engineering.drop_cols.append(item["name"])

        for col in item["cols"][1:]:
            try:
                data[item["name"]] *= data[col]
            except Exception as e:
                warnings.warn("The multiplication operation failed while processing {0}. The error is: {1}".
                              format(col, str(e)), Warning)
                self._conf.feature_engineering.drop_cols.append(item["name"])
        return data

    def addition(self, item, data):
        """Add a sequence of columns to create a new feature.

        :param item: an item in basic_operations_params list
        :param data: the data frame
        :return: the updated data frame
        """
        data[item["name"]] = data[item["cols"][0]]
        if "value" in item:
            try:
                data[item["name"]] = data[item["name"]] + item["value"]
            except Exception as e:
                warnings.warn("The addition operation failed when adding {value} to col {col}. "
                              "The error is: {error}".
                              format(col=item["cols"][0], value=item["value"], error=str(e)), Warning)
                self._conf.feature_engineering.drop_cols.append(item["name"])

        for col in item["cols"][1:]:
            try:
                data[item["name"]] += data[col]
            except Exception as e:
                warnings.warn("The addition operation failed while processing {0}. The error is: {1}".
                              format(col, str(e)), Warning)
                self._conf.feature_engineering.drop_cols.append(item["name"])
        return data

    def subtraction(self, item, data):
        """Subtract a sequence of columns to create a new feature.

        :param item: an item in basic_operations_params list
        :param data: the data frame
        :return: the updated data frame
        """
        data[item["name"]] = data[item["cols"][0]]
        if "value" in item:
            try:
                data[item["name"]] = data[item["name"]] - item["value"]
            except Exception as e:
                warnings.warn("The subtraction operation failed when subtracting {value} from col {col}. "
                              "The error is: {error}".
                              format(col=item["cols"][0], value=item["value"], error=str(e)), Warning)
                self._conf.feature_engineering.drop_cols.append(item["name"])

        for col in item["cols"][1:]:
            try:
                data[item["name"]] -= data[col]
            except Exception as e:
                warnings.warn("The subtraction operation failed while processing {0}. The error is: {1}".
                              format(col, str(e)), Warning)
                self._conf.feature_engineering.drop_cols.append(item["name"])

        return data

    @staticmethod
    def log_transform(data, params):
        """Loop through each item in log_transform_params and apply the log transformation to a specific base defined
        by user. If we end up with inf after the transformation, replace them with nan and handle them later.

        :param data: the train or test dataframe
        :param params: a list of log_transform parameters coming from the conf object (see the configuration module for details)
        :return: the updated dataframe
        """
        for item in params:
            try:
                data[item["name"]] = data[item["col"]] + item["shift"]
                data[item["name"]] = (np.log(data[item["name"]]) / np.log(item["base"])).replace(np.inf, np.nan).replace(-np.inf, np.nan)
            except Exception as e:
                warnings.warn("The log_transform failed while processing {0}. The error is: {1}".
                              format(item["col"], str(e)), Warning)

        return data

    @staticmethod
    def exponential_transform(data, params):
        """Loop through each item in exponential_transform_params and apply the exponential transformation to a
        specific power defined by user. If we end up with inf after the transformation, replace them with nan and
        handle them later.

        :param data: the train or test dataframe
        :param params: a list of exponential_transform parameters coming from the conf object (see the configuration module for details)
        :return: the updated dataframe
        """
        for item in params:
            try:
                data[item["name"]] = np.power(item["power"], data[item["col"]]).replace(np.inf, np.nan).replace(-np.inf, np.nan)
                data[item["name"]] = data[item["name"]] + item["shift"]
            except Exception as e:
                warnings.warn("The exponential transform failed while processing {0}. The error is: {1}".
                              format(item["col"], str(e)), Warning)
        return data

    def groupby(self, data, params):
        """If learner_groupby is None (during training), instantiate a LearnerGroupby object and then call the
        fit_transform method on it. During validation or testing, the learner_groupby would not be None. In that case,
        we call the transform method on it.

        :param data: the train or test dataframe
        :param params: a list of groupby parameters coming from the conf object (see the configuration module for details)
        :return: the updated dataframe
        """
        if not self._learner_groupby:
            self._learner_groupby = LearnerGroupby(params)
            self._learner_groupby.fit_transform(data)
        else:
            self._learner_groupby.transform(data)
        return data

    def reprocess_data(self, data):
        """Once the feature engineering is completed, process the newly created columns to ensure the data is ready for
        training.

        :param data: the data frame
        :return: the updated data
        """
        logging.info("Processing newly created columns after feature engineering, if any...")
        if not self._processor:
            self._processor = DataProcessor(self._conf)

        data = self._processor.handle_to_numeric(data, cols=list(self._conf.feature_engineering.col_dict.keys()))

        self.processor.fill_nan(data,
                                default_value=self._conf.process.fillnan_value,
                                mean_cols=self._conf.feature_engineering.fillnan_mean_cols,
                                median_cols=self._conf.feature_engineering.fillnan_median_cols,
                                mode_cols=self._conf.feature_engineering.fillnan_mode_cols,
                                value_cols=self._conf.feature_engineering.fillnan_value_cols)

        self.processor.drop_cols(data, cols=self._conf.feature_engineering.drop_cols)
        return data

    def handle_feature_engineering(self, data):
        """The main function for calling the appropriate methods for feature engineering depending on the user's input.

        :param data: the data frame
        :return: updated data
        """
        logging.info("Performing feature engineering...")
        if self._conf.feature_engineering.basic_operations_params:
            logging.info("Performing basic_operations feature engineering...")
            data = self.basic_operations(data)

        if self._conf.feature_engineering.log_transform_params:
            logging.info("Performing log_transform feature engineering...")
            data = self.log_transform(data, self._conf.feature_engineering.log_transform_params)

        if self._conf.feature_engineering.exponential_transform_params:
            logging.info("Performing exponential_transform feature engineering...")
            data = self.exponential_transform(data, self._conf.feature_engineering.exponential_transform_params)

        if self._conf.feature_engineering.groupby_params:
            logging.info("Performing groupby feature engineering...")
            data = self.groupby(data, self._conf.feature_engineering.groupby_params)

        data = self.reprocess_data(data)

        return data


class LearnerGroupby:
    """LearnerGroupby is a class for performing the groupby operations on the datasets. The reason for implementing a
     custom class is to be able to use the values from the training data when working with validation or test dataset.
     Basically, we do all the calculations during the training process. We then build a dictionary that maps the values
     of the training data to their corresponding aggregated values. When it comes to validation or test data, we'll use
     that dictionary to set the values. This is commonly overlooked by many people.
     """
    def __init__(self, params):
        """Initialize an object of LearnerGroupby class using the parameters coming from the conf object. These
        parameters have already been validated and updated. We do not perform any validation in this class.

        :param params: a list of groupby parameters coming from the conf object (see the configuration module for details)
        """
        self._params = params
        self._groupby_dict = dict()

    @property
    def groupby_dict(self):
        return self._groupby_dict

    def fit(self, data):
        """If the params has values, loop through the parameters and the aggregation operations to populate the groupby
        dictionary. For each column and aggregation combination, there will be a key in this dictionary. The value for
        each key is another dictionary that maps the original values to their aggregated values. This dictionary will
        then be used to do the transformation.

        :param data: a pandas dataframe. This is usually training dataset.
        :return: None
        """
        if self._params:
            for param in self._params:
                for agg in param["aggregation"]:
                    gb = data.groupby(by=param["col"]).agg({agg["col"]: agg["method"]})
                    key = f"{param['col']}_{agg['col']}_{agg['method']}"
                    self._groupby_dict[key] = dict(zip(gb.index, gb[agg["col"]]))

    def transform(self, data):
        """If the params has values, loop through the parameters and the aggregation operations to create new columns
        using the groupby dictionary. For each column and aggregation combination, there will be a key in this
        dictionary. The value for each key is another dictionary that maps the original values to their aggregated
        values.

        :param data: a pandas dataframe. This is could be train, validation, or test data.
        :return: None
        """
        for param in self._params:
            for agg in param["aggregation"]:
                key = f"{param['col']}_{agg['col']}_{agg['method']}"
                data[agg["name"]] = data[param["col"]].map(self._groupby_dict[key])

    def fit_transform(self, data):
        """Call fit and transform method. This method is typically called during training while the transform method is
        called during testing.

        :param data: a pandas dataframe. This is could be train, validation, or test data.
        :return: None
        """
        self.fit(data)
        self.transform(data)



# Copyright (C) Prizmi, LLC - All Rights Reserved
# Unauthorized copying or use of this file is strictly prohibited and subject to prosecution under applicable laws
# Proprietary and confidential

import getpass
import os
import sys
import logging

from learner.data_worker.data_loader import get_value


class CommunicationConfiguration:
    """Parse the fields in the communication section of the configuration file."""

    def __init__(self, json_config, credentials):
        self._json_config = json_config
        self._credentials = credentials

        self.admin_email_activate = get_value(self._json_config, False, "communication", "admin_params",
                                              "admin_email_params", "activate")
        self.admin_email_address = self.get_admin_email_address()
        self.admin_email_ausername = self.get_username()
        self.admin_email_password = self.get_password()
        self.admin_email_host = get_value(self._json_config, "smtp.gmail.com", "communication", "admin_params",
                                          "admin_email_params", "host")
        self.admin_email_port = get_value(self._json_config, 587, "communication", "admin_params",
                                          "admin_email_params", "port")

    def get_admin_email_address(self):
        if self.admin_email_activate:
            try:
                admin_email_address = self._json_config["communication"]["admin_params"]["admin_email_params"]["email_address"]
                return admin_email_address
            except KeyError:
                logging.critical("admin_email_params was activated but email_address was not provided. Please update "
                                 "your configuration file and try again. Exiting...")
                sys.exit(1)

    def get_username(self):
        if self.admin_email_activate:
            username = os.getenv("LEARNER_ADMIN_EMAIL_ADDRESS")
            if not username:
                try:
                    username = self._json_config["communication"]["admin_params"]["admin_email_params"]["username"]
                except KeyError:
                    try:
                        username = self._credentials["email"]["username"]
                    except (KeyError, TypeError):  # pragma: no cover
                        username = input("Please enter your email address for sending email: ")
            return username

    def get_password(self):
        if self.admin_email_activate:
            password = os.getenv("LEARNER_ADMIN_EMAIL_PASSWORD")
            if not password:
                try:
                    password = self._json_config["communication"]["admin_params"]["admin_email_params"]["password"]
                except KeyError:
                    try:
                        password = self._credentials["email"]["password"]
                    except (KeyError, TypeError):  # pragma: no cover
                        password = getpass.getpass(prompt="Please enter your password for sending emails: ")
            return password



# Copyright (C) Prizmi, LLC - All Rights Reserved
# Unauthorized copying or use of this file is strictly prohibited and subject to prosecution under applicable laws
# Proprietary and confidential

"""The main module for standard engines. A standard engine include classifier and regressor engines."""

import logging

from learner.data_worker.data_manager import TestDataManager, ValidationDataManager
from learner.model_manager.model_initializer import ModelLoader, get_pickle_dir
from learner.model_manager import prediction_manager
from learner.data_worker.output_handler import OutputHandler, pickle_object
from learner.engines.base_standard_engine import BaseStandard
from learner.utilities.timer import timeit
from learner.configuration.configuration import Configuration
from learner.analysis.analysis import Analysis
from learner.utilities.templates import (SAVE_MODEL_PATH, FEATURE_IMPORTANCE_PATH,
                                         SAVE_MODEL_PATH_SEG, FEATURE_IMPORTANCE_PATH_SEG,
                                         LOAD_MODEL_PATH_SEG)


class StandardEngine(BaseStandard):
    """The standard engine class without support for segmentation."""

    def __init__(self, conf: Configuration):
        """Take a conf object (from the Configuration class) to run a standard engine.

        :param conf: a conf object
        """
        super(StandardEngine, self).__init__(conf)
        self._drop_cols = self._conf.column.drop_from_train + [self._conf.column.target_col]

    @timeit("build the model(s)")
    def run_engine(self):
        """The main function for driving the engine.

        :return: None
        """
        if self._conf.model.train_models:
            self.train_models()
            self.predict()
            self.score_prediction()
            self.combine_predictions(self.conf.sep_timetag, self.models_dict)
        else:
            self.load_models(self._conf.workspace.path)

        self.validate_output()
        self.handle_predictions_vs_actuals(self.models_dict or self.loaded_models_dict)

    @timeit("train the model(s)")
    def train_models(self):
        """Train and potentially optimize models based on the user input

        :return: None
        """
        logging.info("Building the models...")
        for tag, mdl in self.models_dict.items():
            logging.info("Fitting %s model", tag)
            # train the model and save it into model dictionary
            # depending on the user input, either do grid_search, cross_validation, or a single fit
            self.fit_models(self.data, mdl, self._drop_cols)

            # save models
            if self._conf.model.save_models:
                filename = SAVE_MODEL_PATH.format(
                    path=self._conf.model.models_dict[tag]["path"],
                    output_name=self._conf.workspace.name,
                    tag=tag,
                    sep_timetag=str(self._conf.sep_timetag),
                    ext=".sav"
                )
                pickle_object(self.conf, self._conf.model.models_dict[tag]["path"], filename, dict({tag: mdl}),
                              self.processor, self.feature_engineering)

            # handle feature importance if requested
            if self._conf.analysis.importance_activate:
                analysis = Analysis(self._conf)
                filename = FEATURE_IMPORTANCE_PATH.format(
                    output_name=self._conf.workspace.name,
                    tag=str(tag),
                    sep_timetag=str(self._conf.sep_timetag)
                )
                analysis.handle_feature_importance(mdl, filename)


class StandardEngineSegment(BaseStandard):
    def __init__(self, conf: Configuration):
        """Take a conf object (from the Configuration class) to run a standard engine with support for segmentation.

        :param conf: a conf object
        """
        super().__init__(conf)
        self._drop_cols = self._conf.column.drop_from_train + [self._conf.column.target_col] + \
            [self._conf.segmenter.test_name] + [self._conf.segmenter.test_col]

    @timeit("build the model(s)")
    def run_engine(self):
        """The main function for driving the engine.

        :return: None
        """
        if self._conf.model.train_models:
            for seg_id, border in enumerate(self._conf.segmenter.seg_list):
                train_data = self.filter_data(self.data, self._conf.segmenter.train_name, border)
                self.train_models(seg_id, border, train_data)
                self.predict(seg_id, border)
        else:
            logging.info("Loading saved model...")
            self.load_models(self._conf.workspace.path)

        if self._conf.workspace.concat_output and self._conf.data.test_prediction_activate:
            output = OutputHandler(self._conf, data_type="test")
            output.concat_save_csv(self._conf.model.models_dict, self._conf.segmenter.seg_list,
                                   self._conf.workspace.name, self._conf.sep_timetag)

        if self._conf.workspace.concat_output and self._conf.data.validation_score_activate:
            output = OutputHandler(self._conf, data_type="validation")
            output.concat_save_csv(self._conf.model.models_dict, self._conf.segmenter.seg_list,
                                   self._conf.workspace.name, self._conf.sep_timetag)

        if self._conf.validation.validate and self._conf.data.test_prediction_activate:
            self.validate_output()

        self.handle_predictions_vs_actuals(self.models_dict or self.loaded_models_dict)
        self.score_prediction()

    @timeit("train the classification model(s)")
    def train_models(self, seg_id, seg_border, train_data):
        """Train and potentially optimize models based on the user input.

        :param seg_id: segment id
        :param seg_border: upper boundary of the segment
        :param train_data: the training data
        :return: None
        """
        logging.info("**************************************************************************")
        logging.info("Training model on segment_id: %i, segment_border: %s", seg_id, seg_border)
        logging.info("There are %i rows of data in this segment", train_data.shape[0])

        for tag, mdls in self.models_dict.items():
            mdl = mdls[seg_id]
            logging.info("Fitting %s model", tag)
            drop_cols = self._drop_cols + [self._conf.segmenter.train_name]
            self.fit_models(train_data, mdl, drop_cols)

            # save models
            if self._conf.model.save_models:
                filename = SAVE_MODEL_PATH_SEG.format(
                    path=mdl["path"],
                    output_name=self._conf.workspace.name,
                    tag=str(tag),
                    seg_id=str(seg_id),
                    sep_timetag=str(self._conf.sep_timetag),
                    ext=".sav"
                )

                pickle_object(self.conf, mdl["path"], filename, dict({tag: mdl}),
                              self.processor, self.feature_engineering)

            # handle feature importance if requested
            if self._conf.analysis.importance_activate:
                analysis = Analysis(self._conf)
                filename = FEATURE_IMPORTANCE_PATH_SEG.format(
                    output_name=self._conf.workspace.name,
                    tag=str(tag),
                    seg_id=str(seg_id),
                    sep_timetag=str(self._conf.sep_timetag)
                )
                analysis.handle_feature_importance(mdl, filename)

    @timeit("load the model(s)")
    def load_models(self, output_path):
        """Load pickled object from path_to_pickle if full directory is provided. Otherwise, it loads the latest pickled
         object.

        :param output_path: directory path to the pickled object
        :return: None
        """
        paths_to_pickle_dir, sample_pickle_files, prefixes = get_pickle_dir(output_path)

        loader = ModelLoader(self._conf)

        nrows_score = 0

        for path_to_pickle_dir, sample_pickle_file, prefix in zip(paths_to_pickle_dir, sample_pickle_files, prefixes):

            self._conf, self.models_dict, self.processor, self.feature_engineering, sep_timetag = loader.load_model(sample_pickle_file)

            for seg_id, seg_border in enumerate(self._conf.segmenter.seg_list):
                logging.info("**************************************************************************")
                logging.info("Making predictions for segment_id: %i, segment_border: %s", seg_id, seg_border)
                filename = LOAD_MODEL_PATH_SEG.format(
                    path=path_to_pickle_dir,
                    prefix=prefix,
                    seg_id=str(seg_id),
                    sep_timetag=str(sep_timetag),
                    ext=".sav"
                )

                self._conf, self.models_dict, self.processor, self.feature_engineering, sep_timetag = loader.load_model(filename)

                tag = next(iter(self.models_dict))
                self.loaded_models_dict[tag] = self.models_dict[tag]

                self.predict(seg_id, seg_border)

                # we need to make sure we add the number of predictions from all the segments
                nrows_score += self._conf.model.nrows_score

        # we need to assign the total number of predictions for later validation
        # we also don't want to double count if we have multiple models
        self._conf.model.nrows_score = nrows_score / len(paths_to_pickle_dir)

    @timeit("make predictions using the trained model(s)")
    def predict(self, seg_id=None, seg_border=None):
        """Call predict_validation if we need to score the models using validation data and call predict_test if we
        need to make predictions on test data.

        :param seg_id: segment id
        :param seg_border: the upper boundary of the segment
        :return: None
        """
        if self._conf.data.validation_score_activate:
            logging.info("Making predictions using validation data")
            self.predict_validation(seg_id, seg_border)
        if self._conf.data.test_prediction_activate:
            logging.info("Making predictions using test data")
            self.predict_test(seg_id, seg_border)

    def predict_validation(self, seg_id=None, seg_border=None):
        """Use the trained models to make prediction on validation data. Unlike test data, the entire validation data
        is loaded in memory when making predictions on validation data.

        :param seg_id: segment id
        :param seg_border: the upper boundary of the segment
        :return: None
        """
        validation_manager = ValidationDataManager(self._conf, self.processor, self.feature_engineering)
        validation_data = self.filter_data(validation_manager.data, self._conf.segmenter.test_name, seg_border)
        for tag, mdls in self.models_dict.items():
            mdl = mdls.get(seg_id, mdls)

            predictor = getattr(prediction_manager, self._conf.engine + "SegmentPredictor")(mdl, self._conf)
            prediction, data = predictor.make_predictions(validation_data,
                                                          data_type="validation",
                                                          column_name=self._conf.data.validation_column_name,
                                                          prediction_type=self._conf.data.validation_prediction_type)
            predictor.write_predictions(tag,
                                        prediction,
                                        data,
                                        mdl=mdl,
                                        seg_id=seg_id,
                                        data_type="validation",
                                        add_version=self._conf.model.version_activate,
                                        version_name=self._conf.model.version_version_name,
                                        version_column_name=self._conf.model.version_column_name
                                        )

    def predict_test(self, seg_id=None, seg_border=None):
        """Use the trained models to make predictions. Making predictions are performed by iterating through the
        test dataset. The test dataset can be very large and we process it in chunks.

        :param seg_id: the seg id
        :param seg_border: the upper boundary of the seg
        :return: None
        """
        test_manager = TestDataManager(self._conf, self.processor, self.feature_engineering)
        reader, indices = test_manager.get_reader_for_segment(self._conf.segmenter.test_col, seg_border)

        num_chunks = 0
        low_limit = 0
        for index, chunk in enumerate(reader):
            num_chunks += 1
            high_limit = low_limit + chunk.shape[0]
            select_rows = indices[(indices < high_limit) & (indices >= low_limit)]
            low_limit += chunk.shape[0]
            chunk = chunk.loc[select_rows, :]

            test_data = test_manager.get_reader_data(chunk)

            for tag, mdls in self.models_dict.items():
                mdl = mdls.get(seg_id, mdls)

                predictor = getattr(prediction_manager, self._conf.engine + "SegmentPredictor")(mdl, self._conf)
                prediction, data = predictor.make_predictions(test_data,
                                                              data_type="test",
                                                              column_name=self._conf.data.test_column_name,
                                                              prediction_type=self._conf.data.test_prediction_type)
                predictor.write_predictions(tag,
                                            prediction,
                                            data,
                                            index,
                                            mdl,
                                            seg_id=seg_id,
                                            data_type="test",
                                            add_version=self._conf.model.version_activate,
                                            version_name=self._conf.model.version_version_name,
                                            version_column_name=self._conf.model.version_column_name
                                            )

            # keep track of number of rows in test dataset for validation purposes
            self.conf.model.nrows_score += chunk.shape[0]

        output = OutputHandler(self._conf, data_type="test")
        output.concat_chunk_csv(num_chunks, self.models_dict, seg_id)



# Copyright (C) Prizmi, LLC - All Rights Reserved
# Unauthorized copying or use of this file is strictly prohibited and subject to prosecution under applicable laws
# Proprietary and confidential

"""Setup the configuration prior to instantiation of a Configuration object. The setup include loading and validating
the json configuration file, configuring the logging setting, and getting a timetag."""
import os
import logging
from datetime import datetime

from learner.validator.input_validator import InputValidator
from learner.setup.logger import Logger, load_json
from learner.validator.input_validator import validate_type


class Setup(Logger, InputValidator):
    """Get the timetag, load and validate the json configuration file and configure the logging settings."""
    def __init__(self, json_config_file):
        """Get a timetag and read the configuration file to validate the configuration file and configure
        logging settings."""
        # get the path and name of the configuration file
        from learner.setup.parser import get_config_file
        config_filepath, config_filename = get_config_file()
        self._json_config_file = json_config_file if json_config_file else config_filepath
        self.timetag = self.get_timetag()
        self._config_filename = config_filename if config_filename else config_filename
        self._json_config = load_json(self._json_config_file)
        self.sep_timetag = self.get_sep_timetag()
        self._schema_file = self.get_schema_file()
        self._json_schema = load_json(self._schema_file)

        # setup logging
        super().__init__(self._json_config, self.sep_timetag)
        # call the parent initializer to validate the configuration file against the schema
        logging.info("---Validation of the configuration file begins---")
        super(Logger, self).__init__(self._json_config, self._json_schema)

    @property
    def json_config_file(self):
        return self._json_config_file

    @property
    def json_config(self):
        return self._json_config

    @property
    def config_filename(self):
        return self._config_filename

    @property
    def schema_file(self):
        return self._schema_file

    @property
    def json_schema(self):
        return self._json_schema

    def get_schema_file(self):
        try:
            return self._json_config["schema_file"]
        except KeyError:  # pragma: no cover
            default_schema = os.path.join(os.path.dirname(__file__),
                                          "..",
                                          "schema",
                                          "schema.json")
            return default_schema

    @staticmethod
    def get_timetag(date_time=datetime.now()):  # pragma: no cover

        return date_time.strftime("%Y-%m-%d-%H-%M")

    def get_sep_timetag(self):
        try:
            add_timetag = self._json_config["workspace"]["add_timetag"]
            validate_type(add_timetag, bool, object_name="add_timetag", type_name="boolean")
            if not add_timetag:
                return ""
            return f"_{self.timetag}"
        except KeyError:
            return f"_{self.timetag}"



# Copyright (C) Prizmi, LLC - All Rights Reserved
# Unauthorized copying or use of this file is strictly prohibited and subject to prosecution under applicable laws
# Proprietary and confidential

import sys
import copy
import warnings
import logging

from learner.data_worker.data_loader import get_dtype_dict as data_worker_get_dtype_dict
from learner.data_worker.data_loader import get_data, get_value


class ColumnConfiguration:
    """Parse all input variables related to selecting the columns of the datasets including the columns that must be
    used (use_cols), target column (target_col), etc."""

    def __init__(self, json_config, engine, data, sample):
        self._data = data
        self._sample = sample
        self._json_config = json_config
        self._engine = engine
        self.use_cols = self.get_use_cols()
        self.valid_cols = copy.deepcopy(self.use_cols)
        self.target_col = self.get_target_col()
        self.path_col = self.get_path_col()
        self.copy_cols = self.get_copy_cols()
        self.all_cols = self.get_all_cols()
        self.col_dtypes = get_value(self._json_config, None, "column", "dtype_params", "col_dtypes")

        self.dtype_activate = self.get_dtypes_activate()
        self.dtype_trial_nrows = self.get_dtype_trial_nrows()
        self.dtype_dict = self.get_dtype_dict()

        # these happen here so that we can pass the new column around and avoid validations errors
        self.date_cols_activate = get_value(self._json_config, False, "process", "date_cols_params", "activate")
        self.col_names_for_date_cols_items = self.get_col_names_for_date_cols_items()

        self.id_cols = self.get_id_cols()
        self.drop_cols = self.get_drop_cols()
        self.drop_from_train = self.get_drop_from_train()

    def get_use_cols(self):
        try:
            return self._json_config["column"]["use_cols"]
        except KeyError:
            if self._data.train_input_type == "file":
                warnings.warn("No use_cols has been specified in json configuration. Using all columns from the train "
                              "dataset.", UserWarning)
                use_cols = get_data(self._data.train_location,
                                    self._data.manifest,
                                    format=self._data.train_format,
                                    nrows=1,
                                    sep=self._data.train_delimiter,
                                    header=self._data.train_header,
                                    usecols=None).columns.tolist()
                return use_cols
            return []

    def get_copy_cols(self):
        try:
            copy_cols = self._json_config["column"]["copy_cols"]
            # we don't allow copying the target column because we don't load target when testing
            if any([col == self.target_col for col in copy_cols]):
                logging.critical("""
                The target columns cannot be included in copy_cols in the column section. Please update 
                your configuration file and retry. Exiting...
                """)
                sys.exit(1)

            # add the new columns to valid column names
            self.valid_cols.extend(list(copy_cols.values()))
            return copy_cols
        except KeyError:
            return {}

    def get_all_cols(self):
        if self._data.manifest:
            all_cols = get_data(self._data.train_location,
                                self._data.manifest,
                                format=self._data.train_format,
                                nrows=1,
                                sep=self._data.train_delimiter,
                                header=self._data.train_header,
                                usecols=None).columns.tolist()
            return all_cols

        return None

    def get_dtypes_activate(self):
        try:
            activate = self._json_config["column"]["dtype_params"]["activate"]
            return activate
        except KeyError:
            if self.col_dtypes:
                warnings.warn("col_dtypes was specified but load_with_dtypes was not set to true. Loading with "
                              "dtype won't be activated.", UserWarning)
            return False

    def get_dtype_trial_nrows(self):
        recommended_nrows = 200000
        try:
            dtype_trial_nrows = self._json_config["column"]["dtype_params"]["nrows"]
            if self.dtype_activate:
                if dtype_trial_nrows < recommended_nrows and self._data.train_nrows and dtype_trial_nrows < self._data.train_nrows:
                    warnings.warn("It is recommended to use {0} rows for trial loading but {1} rows was requested."
                                  .format(recommended_nrows, dtype_trial_nrows), UserWarning)
                    return dtype_trial_nrows
                if self._data.train_nrows and dtype_trial_nrows > self._data.train_nrows:
                    warnings.warn("{0} rows was requested for trial loading but {1} is used for training. Setting "
                                  "trial nrows to {1}".format(dtype_trial_nrows, self._data.train_nrows)
                                  .format(recommended_nrows, dtype_trial_nrows), UserWarning)
                    return self._data.train_nrows
        except KeyError:
            if self._data.train_nrows and self._data.train_nrows < recommended_nrows:
                return self._data.train_nrows
            return recommended_nrows

    def get_dtype_dict(self):
        if self.dtype_activate:
            dtype_dict = data_worker_get_dtype_dict(self._data.meta_data, self.use_cols, self.col_dtypes)
            return dtype_dict

    def get_target_col(self):
        try:
            target_col = self._json_config["column"]["target_col"]
            # if train_test_split is active, we don't need to know if the target column exists in validation_cols or not
            # because we will use the train data which would definitely include the target
            # also if the query in validation query is activated, we don't need to get the validation columns
            # finally, we only need to get the columns if the input_type is "file"
            if self._data.validation_score_activate and not self._sample.split_activate and not self._data.validation_query_activate and self._data.validation_input_type == "file":
                # we only do this validation if we have the csv format. For other formats, we'd have to load the
                # entire data, which could slow things down. If there's an issue, it will surface later
                if self._data.validation_format == "csv":
                    all_validation_cols = self.get_all_validation_cols()
                    if target_col not in all_validation_cols:
                        logging.error("score_predictions is set to true but target_col is not in validation data. "
                                      "Exiting...")
                        sys.exit(1)
                # if the input_type is not file, we can simply add the target_col to valid_cols
            if (self._data.train_input_type != "file" or self._data.validation_input_type != "file") and target_col not in self.valid_cols:
                self.valid_cols.append(target_col)
            return target_col
        except KeyError:
            if self._data.validation_input_type == "file":
                logging.critical("target_col cannot be undefined. Exiting...")
                sys.exit(1)
            target_col = "target"
            self.valid_cols.append(target_col)
            return target_col

    def get_path_col(self):
        if self._engine == "ImageClassifier":
            path_col = "path"
            try:
                path_col = self._json_config["column"]["path_col"]
            finally:
                self.valid_cols.append(path_col)
                return path_col

    def get_all_validation_cols(self):
        cols = get_data(self._data.validation_location,
                        self._data.manifest,
                        nrows=1,
                        sep=self._data.validation_delimiter,
                        header=self._data.validation_header,
                        usecols=None).columns.tolist()
        return cols

    def get_col_names_for_date_cols_items(self):
        try:
            date_cols = self._json_config["process"]["date_cols_params"]["date_cols"]
            col_names_for_date_cols_items = []
            if self.date_cols_activate and date_cols:
                for col, date_item in date_cols.items():
                    # make sure the user can make requests on the new columns after they've been created
                    for item in date_item:
                        col_names_for_date_cols_items.append("{col}_{item}".format(col=col, item=item))

            self.valid_cols.extend(col_names_for_date_cols_items)
            return col_names_for_date_cols_items
        except KeyError:
            return []

    def get_id_cols(self):
        try:
            id_cols = self._json_config["column"]["id_cols"]
            return id_cols
        except KeyError:
            return []

    def get_drop_cols(self):
        try:
            drop_cols = self._json_config["column"]["drop_cols"]
            return drop_cols
        except KeyError:
            return []

    def get_drop_from_train(self):
        try:
            drop_from_train = self._json_config["column"]["drop_from_train"]
            return drop_from_train
        except KeyError:
            return []



# Copyright (C) Prizmi, LLC - All Rights Reserved
# Unauthorized copying or use of this file is strictly prohibited and subject to prosecution under applicable laws
# Proprietary and confidential

import os
import warnings
import logging

from learner.engines.engine_manager import EngineHandler
from learner.configuration.configuration import Configuration
from learner.communication.communication_manager import CommunicationHandler
from learner.data_worker.data_mover import DataMoverManager
os.environ["JOBLIB_START_METHOD"] = "forkserver"
warnings.filterwarnings(action='ignore', category=DeprecationWarning)


def main():
    conf = None
    try:
        conf = Configuration()
        # we handle s3 right after we create the conf object
        data_mover = DataMoverManager(conf)
        data_mover.move_data(mode="download")
        EngineHandler(conf)
        data_mover.move_data(mode="upload")
        message = "{timetag}. SUCCESS. CONFIG: {config_filename}".format(timetag=conf.timetag,
                                                                         config_filename=conf.config_filename)
        comm = CommunicationHandler(conf, message=message)
        comm.handle_admin_communication()
    except Exception as e:
        # we still want to upload to s3 (if requested) even if we get an error
        data_mover = DataMoverManager(conf)
        data_mover.move_data(mode="upload")
        try:
            message = "{timetag}. ERROR: '{error}'. CONFIG: {config_filename}.".format(timetag=conf.timetag,
                                                                                       config_filename=conf.config_filename,
                                                                                       error=str(e))
            logging.critical(f"Something went wrong during execution. Here's what we know about the error: {e}")
            comm = CommunicationHandler(conf, message=message)
            comm.handle_admin_communication()
        except AttributeError:
            # if we get here, that means we don't have a conf object. We need to get the info we need and then
            # communicate them.
            logging.critical("Problem in parsing the configuration file. The error is: {error}".format(error=str(e)))
